"""
Some constants, mappings, and functions to be used across the software
"""

from __future__ import annotations
from itertools import chain
import os
from multiprocessing import Pool
import ads
from ads.exceptions import APIResponseError
import json
import astropy.units as u
import numpy as np

"""
Helper functions first that just don't belong anywhere else
"""


def filter_to_obstype(band_name):
    """
    Converts a band name to either 'radio', 'uvoir', 'xray'
    """

    try:
        wave_eff = FILTER_MAP_WAVE[band_name] * u.nm
    except KeyError as exc:
        raise Exception(
            f"No Effective Wavelength Known for {band_name}, please add it to constants"
        ) from exc

    return wave_to_obstype(wave_eff)


def freq_to_obstype(freq_eff):
    """
    Converts a frequency value to either 'radio', 'uvoir', 'xray'

    Args:
        freq_eff (u.Quantity) : An astropy quantity in frequency space
    """

    wave_eff = freq_eff.to(u.nm, equivalencies=u.spectral())
    return wave_to_obstype(wave_eff)


def wave_to_obstype(wave_eff):
    """
    Converts a wavelength value to either 'radio', 'uvoir', 'xray'

    Args:
        wave_eff (u.Quantity) : An astropy quantity in wavelength space
    """
    if wave_eff > 0.1 * u.mm:
        return "radio"
    elif wave_eff <= 0.1 * u.mm and wave_eff >= 10 * u.nm:
        return "uvoir"
    else:
        return "xray"


def clean_schema(schema):
    """
    Clean out Nones and empty lists from the given subschema
    """
    for key, val in list(schema.items()):
        if val is None or (isinstance(val, (list, dict)) and len(val) == 0):
            del schema[key]
    return schema


def bibcode_to_hrn(bibcode, local_reference_map="reference_map_local.json"):
    if isinstance(bibcode, str):
        bibcode = [bibcode]

    if os.path.exists(local_reference_map):
        with open(local_reference_map, "r") as j:
            local_map = json.load(j)
    else:
        local_map = {}

    hrns = []
    bibcodes = []
    bibcodes_to_query = []
    for b in bibcode:
        if b in local_map:
            hrns.append(local_map[b])
            bibcodes.append(b)
        else:
            bibcodes_to_query.append(b)

    if len(bibcodes_to_query) > 0:
        queried_bibs, queried_hrns = _bibcode_to_hrn_with_query(bibcodes_to_query)
        hrns += queried_hrns
        bibcodes += queried_bibs

    return bibcodes, hrns


def _bibcode_to_hrn_with_query(bibcode):
    """
    Converts a bibcode to a human_readable_name (hrn) using ADSQuery
    """
    if isinstance(bibcode, str):
        bibcode = bibcode.strip()
        query = f"bibcode:{bibcode}"
        bibcodes = [bibcode]
    else:
        # make sure the bibcodes are lists instead of strings of lists
        bibcode = [b.strip("[]").replace("'", "").split(", ") for b in bibcode]

        bibcode = list(chain(*[v if isinstance(v, list) else [v] for v in bibcode]))

        bibcodes_flat = np.array(bibcode).flatten()
        bibcodes_cleaned = np.array([b.strip() for b in bibcodes_flat])

        bibcodes = list(np.unique(bibcodes_cleaned))

    protected_vals = ["private", "new", "current work"]
    for val in protected_vals:
        if val in bibcodes:
            bibcodes.pop(bibcodes.index(val))

    query = f"bibcode:{bibcodes[0]}"
    if len(bibcodes) > 1:
        for b in bibcodes[1:]:
            query += f" OR {b}"

    try:
        qobj = ads.SearchQuery(q=query)
        qobj.execute()  # do the query
        adsquery = list(qobj)
    except IndexError:
        raise ValueError(f"Could not find {bibcode} on ADS!")
    except APIResponseError as exc:
        raise ValueError(
            f"""Failed on query {query}! \n Potentially out of ADS
            queries! Run curl command to check like\n
            https://github.com/adsabs/adsabs-dev-api/blob/master/README.md"""
        ) from exc

    hrns = []
    bibcodes_to_return = []
    for res in adsquery:
        authors = res.author
        year = res.year

        if len(authors) == 0:
            raise ValueError("This ADS bibcode does not exist!")
        elif len(authors) == 1:
            author = authors[0]
        elif len(authors) == 2:
            author = authors[0] + " & " + authors[1]
        else:  # longer than 2
            author = authors[0] + " et al."

        # generate the human readable name
        hrn = author + " (" + year + ")"
        hrns.append(hrn)
        bibcodes_to_return.append(res.bibcode)

    if len(hrns) == 0 and len(bibcodes) != 0:
        raise ValueError(f"Could not find any bibcodes associated with {bibcodes}!")

    if isinstance(bibcode, str):
        return hrns[0]

    if len(hrns) != len(bibcodes):
        raise ValueError(
            f"ADS has multiple sources associated with one of these bibcodes! \
            {bibcode}"
        )

    return bibcodes_to_return, hrns


def freq_to_band(freq: u.Quantity) -> str:
    """
    Converts an effective frequency to the corresponding band name based on the
    standards listed in RADIO_BAND_MAPPING

    Args:
        freq (astropy Quantity) : Astropy Quantity of the frequency to get the band name
    Returns:
        string with the commonly used band name
    """

    for key, freq_range in RADIO_BAND_MAPPING.items():
        if freq_range[0] * u.GHz < freq <= freq_range[1] * u.GHz:
            return key

    raise ValueError(f"No band name found for the frequency {freq}. Please verify that \
                      it is in a range in RADIO_BAND_MAPPING and, if not, add it!")


def _wrap_single_conversion(t):
    """
    t is a tuple of (frequency, unit)
    """
    return freq_to_band(t[0] * u.Unit(t[1]))


def freqlist_to_band(
    freq_list: list[float], freq_unit_list: list[str], ncores=1
) -> list[str]:
    """
    Converts a list of effective frequencies to the corresponding band names based on
    the standards listed in RADIO_BAND_MAPPING

    Args:
        freq_list (list[float]): floats for the frequencies
        freq_unit_list (list[str]): List of astropy unit strings to apply to freq_list
        ncores (int): The number of cores to multiprocess with

    Returns:
        list of strings with the band names
    """
    with Pool(ncores) as p:
        return p.map(_wrap_single_conversion, zip(freq_list, freq_unit_list))


"""
Then the constants and dictionary mappings used throughout
"""

# gives the effective wavelength for each filter given
# these are all in nanometers!
FILTER_MAP_WAVE = {
    "FUV": 153.8620701901866,
    "NUV": 231.56631043707714,
    "UVW2": 207.98754332676123,
    "uvw2": 207.98754332676123,
    "W2": 207.98754332676123,
    "2": 207.98754332676123,
    "uw2": 207.98754332676123,
    "UVM2": 225.47802478793594,
    "uvm2": 225.47802478793594,
    "M2": 225.47802478793594,
    "M": 225.47802478793594,
    "um2": 225.47802478793594,
    "UVW1": 261.3713060531025,
    "uvw1": 261.3713060531025,
    "W1": 261.3713060531025,
    "1": 261.3713060531025,
    "uw1": 261.3713060531025,
    "u": 356.17887353001856,
    "u'": 356.17887353001856,
    "up": 356.17887353001856,
    "uprime": 356.17887353001856,
    "U_S": 347.06360491031495,
    "s": 347.06360491031495,
    "us": 347.06360491031495,
    "U": 353.10502283105023,
    "B": 443.0539845758355,
    "B_S": 435.912081730874,
    "b": 435.912081730874,
    "bs": 435.912081730874,
    "g": 471.8872246248687,
    "g'": 471.8872246248687,
    "gp": 471.8872246248687,
    "gprime": 471.8872246248687,
    "F475W": 471.8872246248687,
    "g-DECam": 482.6787274749997,
    "c": 540.8724658332794,
    "cyan": 540.8724658332794,
    "V": 553.7155963302753,
    "V_S": 543.0131091205997,
    "v": 543.0131091205997,
    "vs": 543.0131091205997,
    "Itagaki": 651.0535687558726,
    "white": 752.0,
    "unfilt.": 616.690135,
    "0": 616.690135,
    "C": 616.690135,
    "clear": 616.690135,
    "pseudobolometric": 616.690135,
    "griz": 616.690135,
    "RGB": 616.690135,
    "LGRB": 616.690135,
    "G": 673.5412573108297,
    "Kepler": 641.6835660569259,
    "TESS": 797.2360657697333,
    "DLT40": 615.8130149792426,
    "Open": 615.8130149792426,
    "Clear": 615.8130149792426,
    "w": 638.9300625093831,
    "o": 686.6260690394873,
    "orange": 686.6260690394873,
    "r": 618.5194476741524,
    "r'": 618.5194476741524,
    "rp": 618.5194476741524,
    "rprime": 618.5194476741524,
    "F625W": 618.5194476741524,
    "r-DECam": 643.2062638192127,
    "R": 646.9439215118385,
    "Rc": 646.9439215118385,
    "R_s": 646.9439215118385,
    "i": 749.9704174464691,
    "i'": 749.9704174464691,
    "ip": 749.9704174464691,
    "iprime": 749.9704174464691,
    "F775W": 749.9704174464691,
    "i-DECam": 782.6680306208917,
    "I": 788.558706467662,
    "Ic": 788.558706467662,
    "z_s": 867.9495480864285,
    "zs": 867.9495480864285,
    "z": 896.1488333992431,
    "z'": 896.1488333992431,
    "zp": 896.1488333992431,
    "zprime": 896.1488333992431,
    "z-DECam": 917.8949537472383,
    "y": 963.3308299506817,
    "y-DECam": 989.965087304703,
    "J": 1255.0918319447906,
    "H": 1630.5155019191195,
    "K": 2157.3600605745955,
    "Ks": 2157.3600605745955,
    "F070W": 705.5727879998312,
    "F090W": 904.2281265089156,
    "F115W": 1157.001589027877,
    "F150W": 1503.9880463410511,
    "F200W": 1993.3922957570885,
    "F225W": 2372.81,
    "F277W": 2769.332372846113,
    "F300M": 2990.7606605760484,
    "F335M": 3363.887076210947,
    "F356W": 3576.787256375927,
    "F360M": 3626.0578695682693,
    "F444W": 4415.974447587756,
    "F560W": 5645.279496731566,
    "F770W": 7663.455798629626,
    "F1000W": 9968.161727011531,
    "F1130W": 11310.984595876938,
    "F1280W": 12831.396996921212,
    "F1500W": 15091.367399905488,
    "F1800W": 18006.083119653664,
    "F2100W": 20842.526633138932,
    "F2550W": 25408.228367890282,
}
"""
Mapping for the effective wavelength in nanometers for all filters used in the dataset.
"""


# gives the effective frequency for all filters
# These are all in THz
FILTER_MAP_FREQ = {
    "FUV": 1975.086895569116,
    "NUV": 1346.3548820463916,
    "UVW2": 1531.4976984760474,
    "uvw2": 1531.4976984760474,
    "W2": 1531.4976984760474,
    "2": 1531.4976984760474,
    "uw2": 1531.4976984760474,
    "UVM2": 1360.083095675749,
    "uvm2": 1360.083095675749,
    "M2": 1360.083095675749,
    "M": 1360.083095675749,
    "um2": 1360.083095675749,
    "UVW1": 1236.8527545450256,
    "uvw1": 1236.8527545450256,
    "W1": 1236.8527545450256,
    "1": 1236.8527545450256,
    "uw1": 1236.8527545450256,
    "u": 849.2871562331687,
    "u'": 849.2871562331687,
    "up": 849.2871562331687,
    "uprime": 849.2871562331687,
    "U_S": 875.611103788721,
    "s": 875.611103788721,
    "us": 875.611103788721,
    "U": 858.321721875779,
    "B": 688.8500955332158,
    "B_S": 696.7876979144597,
    "b": 696.7876979144597,
    "bs": 696.7876979144597,
    "g": 648.9823425403824,
    "g'": 648.9823425403824,
    "gp": 648.9823425403824,
    "gprime": 648.9823425403824,
    "F475W": 648.9823425403824,
    "g-DECam": 635.8015668464043,
    "c": 580.1132515050684,
    "cyan": 580.1132515050684,
    "V": 548.3068934496129,
    "V_S": 554.9815375506427,
    "v": 554.9815375506427,
    "vs": 554.9815375506427,
    "Itagaki": 577.0861573682259,
    "white": 30079.243284322874,
    "unfilt.": 601.5655810567023,
    "0": 601.5655810567023,
    "C": 601.5655810567023,
    "clear": 601.5655810567023,
    "pseudobolometric": 601.5655810567023,
    "griz": 601.5655810567023,
    "RGB": 601.5655810567023,
    "LGRB": 601.5655810567023,
    "G": 518.6766845466752,
    "Kepler": 519.5058608954615,
    "TESS": 403.1881955125893,
    "DLT40": 629.637672549936,
    "Open": 629.637672549936,
    "Clear": 629.637672549936,
    "w": 520.8387777057242,
    "o": 451.71177203298663,
    "orange": 451.71177203298663,
    "r": 489.2629992899134,
    "r'": 489.2629992899134,
    "rp": 489.2629992899134,
    "rprime": 489.2629992899134,
    "F625W": 489.2629992899134,
    "r-DECam": 472.4459671948087,
    "R": 471.26223689126897,
    "Rc": 471.26223689126897,
    "R_s": 471.26223689126897,
    "i": 402.8409598867557,
    "i'": 402.8409598867557,
    "ip": 402.8409598867557,
    "iprime": 402.8409598867557,
    "F775W": 402.8409598867557,
    "i-DECam": 386.62233825433924,
    "I": 382.7915178046724,
    "Ic": 382.7915178046724,
    "z_s": 346.66628641927826,
    "zs": 346.66628641927826,
    "z": 337.7343708777923,
    "z'": 337.7343708777923,
    "zp": 337.7343708777923,
    "zprime": 337.7343708777923,
    "z-DECam": 328.753462451287,
    "y": 312.24818210606065,
    "y-DECam": 303.4727730182509,
    "J": 239.862442505934,
    "H": 185.33613196897403,
    "K": 139.79431978859097,
    "Ks": 139.79431978859097,
    "F070W": 431.04176743403116,
    "F090W": 336.17431986268366,
    "F115W": 262.87628654288676,
    "F150W": 201.94374815011136,
    "F200W": 152.56522352568953,
    "F277W": 110.05136786468209,
    "F300M": 100.56915203596012,
    "F335M": 89.41072625742719,
    "F356W": 85.01984846997881,
    "F360M": 82.9357933095218,
    "F444W": 68.96667222373961,
    "F560W": 53.67852315133938,
    "F770W": 39.87175477126777,
    "F1000W": 30.349460503852665,
    "F1130W": 26.53952983680919,
    "F1280W": 23.59741975845449,
    "F1500W": 20.08679352819493,
    "F1800W": 16.773842151606242,
    "F2100W": 14.581938602646188,
    "F2550W": 11.919267708332558,
}
"""
Mapping for the effective frequencies in THz for all the filters used in OTTER
"""


# x-ray telescope areas for converting
# NOTE: these are estimates from the links provided
# Since this is inherently instrument dependent they are not entirely reliable
# All are for 1-2 keV
XRAY_AREAS = {
    # https://swift.gsfc.nasa.gov/about_swift/Sci_Fact_Sheet.pdf
    "swift": 135 * u.cm**2,
    # https://heasarc.gsfc.nasa.gov/docs/rosat/ruh/handbook/node39.html#SECTION00634000000000000000
    "rosat": 400 * u.cm**2,
    # https://www.cosmos.esa.int/web/xmm-newton/technical-details-mirrors
    "xmm": 1500 * u.cm**2,
    "xmm slew": 1500 * u.cm**2,
    "xmm-slew": 1500 * u.cm**2,
    "xmm pointed": 1500 * u.cm**2,
    "xmm-newton": 1500 * u.cm**2,
    # https://cxc.harvard.edu/cdo/about_chandra
    "chandra": 600 * u.cm**2,
    # https://www.cosmos.esa.int/documents/332006/954765/Brunner_TopicK.pdf
    "erosita": 1500 * u.cm**2,
}
"""
X-Ray telescope areas that are used for converting from counts to other units.

NOTE: These are estimates from the following links
* https://swift.gsfc.nasa.gov/about_swift/Sci_Fact_Sheet.pdf
* https://heasarc.gsfc.nasa.gov/docs/rosat/ruh/handbook/node39.html#SECTION00634000000000000000
* https://www.cosmos.esa.int/web/xmm-newton/technical-details-mirrors
* https://cxc.harvard.edu/cdo/about_chandra
"""

# Radio Band names with frequency ranges in GHz
RADIO_BAND_MAPPING = {
    # use some lower bands from IEEE standards up to GMRT standards
    # These are mostly here just as catchalls for low frequency radio telescopes
    # like GEETEE or LOFAR
    "HF": (0.003, 0.03),
    "VHF": (0.03, 0.125),  # Official: (0.03, 0.3) but adjust to work with GMRT standard
    # use GMRT standards for low frequencies
    "gmrt.2": (0.125, 0.250),
    "gmrt.3": (0.25, 0.55),  # Official: (0.25, 0.5)
    # IEEE Standard Naming
    # https://terasense.com/terahertz-technology/radio-frequency-bands/
    "UHF": (0.55, 1),
    "L": (1, 2),
    "S": (2, 4),
    "C": (4, 8),
    "X": (8, 12),
    "Ku": (12, 18),
    "K": (18, 27),
    "Ka": (27, 40),
    # then switch to ALMA standard for mm
    # https://www.eso.org/public/teles-instr/alma/receiver-bands/
    # I widened some of the official ranges slightly so we don't leave gaps
    # although these gaps probably wouldn't have mattered because **atmosphere**
    "alma.1": (40, 50),  # Official: (35, 50)
    "alma.3": (84, 116),  # skip alma.2 cause it's just a smaller range of alma.3
    "alma.4": (116, 163),  # official: (125, 163)
    "alma.5": (163, 211),
    "alma.6": (211, 275),
    "alma.7": (275, 373),
    "alma.8": (373, 500),  # official: (385,500)
    "alma.9": (500, 720),  # official: (602, 720)
    "alma.10": (787, 950),
}
"""
Mapping of common radio/mm band names to their corresponding frequency ranges.
All frequencies are in GHz. The upperlimit is inclusive, that is these ranges
are really (xx, yy].
"""

# define a working base directory constant
BASEDIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
"""
Base directory for the OTTER API software package
"""

VIZIER_LARGE_CATALOGS = [
    "2MASS-PSC",
    "2MASX",
    "AC2000.2",
    "AKARI",
    "ALLWISE",
    "ASCC-2.5",
    "B/DENIS",
    "CMC14",
    "Gaia-DR1",
    "GALEX",
    "GLIMPSE",
    "GSC-ACT",
    "GSC1.2",
    "GSC2.2",
    "GSC2.3",
    "HIP",
    "HIP2",
    "IRAS",
    "NOMAD1",
    "NVSS",
    "PanSTARRS-DR1",
    "PGC",
    "Planck-DR1",
    "PPMX",
    "PPMXL",
    "SDSS-DR12",
    "SDSS-DR7",
    "SDSS-DR9",
    "Tycho-2",
    "UCAC2",
    "UCAC3",
    "UCAC4",
    "UKIDSS",
    "USNO-A2",
    "USNO-B1",
    "WISE",
]
"""
ViZier catalog names that we query for host information in the Host class
"""

_KNOWN_CLASS_ROOTS = [
    "SN",
    "SLSN",
    "TDE",
    "GRB",
    "LGRB",
    "SGRB",
    "AGN",
    "FRB",
    "QSO",
    "ANT",
]
"""
Classification root names
"""

DATADIR = os.path.join(BASEDIR, "data", "base")
"""
Deprecated database directory that IS NOT always constant anymore
"""

####################################################################
# The following schema dictionaries are here so the code remains
# backwards compatible. BUT! All future code should use the pydantic
# schema defined in schema.py
####################################################################

# Overarching schema that stops once we get down to a string or list
schema = {
    "schema_version": {"value": "0", "comment": "Copied from tde.space"},
    "name": {"default_name": None, "alias": []},
    "coordinate": [],
    "distance": [],
    "classification": {"value": []},
    "reference_alias": [],
    "date_reference": [],
    "photometry": [],
    "spectra": [],
    "filter_alias": [],
}
"""
Schema dictionary to be filled with values from the subschemas
"""

# sub schemas that get filled into lists
name_alias_schema = {"value": None, "reference": None}
"""
Subschema for the name and alias dictionary
"""

coordinate_schema = {
    "ra": None,
    "dec": None,
    "l": None,
    "b": None,
    "lon": None,
    "lat": None,
    "ra_units": None,
    "dec_units": None,
    "l_units": None,
    "b_units": None,
    "lon_units": None,
    "lat_units": None,
    "ra_error": None,
    "dec_error": None,
    "l_error": None,
    "b_error": None,
    "lon_error": None,
    "lat_error": None,
    "epoch": None,
    "frame": None,
    "coord_type": None,
    "computed": None,
    "reference": None,
    "default": None,
}
"""
Subschema to describe the possible attributes for the coordinate dictionary
"""


distance_schema = {
    "value": None,
    "unit": None,
    "error": None,
    "cosmology": None,
    "reference": None,
    "computed": None,
    "uuid": None,
    "default": None,
    "distance_type": None,
}
"""
Subschema to describe the possible attributes for the distance dictionary
"""


classification_schema = {
    "object_class": None,
    "confidence": None,
    "class_type": None,
    "reference": None,
    "default": None,
}
"""
Subschema to describe the attributes for the classification dictionary
"""

reference_alias_schema = {"name": None, "human_readable_name": None}
"""
Subschema to describe the attributes for the reference alias dictionary
"""

date_reference_schema = {
    "value": None,
    "date_format": None,
    "date_type": None,
    "reference": None,
    "computed": None,
}
"""
Subschema to describe the date_reference dictionary attributes
"""

photometry_schema = {
    "raw": None,
    "raw_err": None,
    "raw_units": None,
    "value": None,
    "value_err": None,
    "value_units": None,
    "epoch_zeropoint": None,
    "epoch_redshift": None,
    "filter": None,
    "filter_key": None,
    "obs_type": None,
    "telescope_area": None,
    "date": None,
    "date_format": None,
    "date_err": None,
    "ignore": None,
    "upperlimit": None,
    "sigma": None,
    "sky": None,
    "telescope": None,
    "instrument": None,
    "phot_type": None,
    "exptime": None,
    "aperature": None,
    "observer": None,
    "reducer": None,
    "pipeline": None,
    "corr_k": None,
    "corr_av": None,
    "corr_host": None,
    "corr_hostav": None,
    "val_k": None,
    "val_s": None,
    "val_av": None,
    "val_host": None,
    "val_hostav": None,
}
"""
Subschema to describe all of the possible attributes that can be used in the photometry
dictionary
"""

spectra_schema = {
    "wavelength": None,
    "wavelength_units": None,
    "flux": None,
    "fluxerr": None,
    "raw": None,
    "raw_err": None,
    "sky": None,
    "lamp": None,
    "flux_units": None,
    "telescope": None,
    "instrument": None,
    "date": None,
    "date_format": None,
    "date_err": None,
    "exptime": None,
    "slit": None,
    "airmass": None,
    "disperser": None,
    "resolution": None,
    "resolution_units": None,
    "min_wave": None,
    "max_wave": None,
    "filter": None,
    "filter_key": None,
    "standard_name": None,
    "ignore": None,
    "spec_type": None,
    "aperture": None,
    "observer": None,
    "reducer": None,
    "pipeline": None,
    "corr_k": None,
    "corr_av": None,
    "corr_host": None,
    "corr_hostav": None,
    "corr_flux": None,
    "corr_phot": None,
    "val_k": None,
    "val_av": None,
    "val_host": None,
    "val_hostav": None,
}

filter_alias_schema = {
    "filter_key": None,
    "wave_eff": None,
    "wave_min": None,
    "wave_max": None,
    "freq_eff": None,
    "freq_min": None,
    "freq_max": None,
    "zp": None,
    "wave_units": None,
    "freq_units": None,
    "zp_units": None,
    "zp_system": None,
}
"""
Subschema to describe the attributes in the filter_alias dictionary
"""

# package the subschemas by the key used for that location in the Transient object
subschema = {
    "name/alias": name_alias_schema,
    "coordinate": coordinate_schema,
    "distance": distance_schema,
    "classification": classification_schema,
    "reference_alias": reference_alias_schema,
    "date_reference": date_reference_schema,
    "photometry": photometry_schema,
    "spectra": spectra_schema,
    "filter_alias": filter_alias_schema,
}
"""
A useful variable to describe all of the subschemas that are available and can be used
"""


class _DuplicateFilter(object):
    def __init__(self):
        self.msgs = set()

    def filter(self, record):
        rv = record.msg not in self.msgs
        self.msgs.add(record.msg)
        return rv



"""
Make sure all of the exceptions don't throw silly errors
"""

import pytest
from otter import exceptions as exc


def test_failed_query_error():
    with pytest.raises(exc.FailedQueryError):
        raise exc.FailedQueryError()

    txt = "You're query/search did not return any results! "
    txt += "Try again with different parameters!"

    assert str(exc.FailedQueryError()) == txt


def test_ioerror():
    with pytest.raises(exc.IOError):
        raise exc.IOError


def test_otter_limitation_error():
    with pytest.raises(exc.OtterLimitationError):
        raise exc.OtterLimitationError("")

    assert str(exc.OtterLimitationError("foo")) == "Current Limitation Found: foo"


def test_transient_merge_error():
    with pytest.raises(exc.TransientMergeError):
        raise exc.TransientMergeError


def test_otter_not_implemented_error():
    with pytest.raises(exc.OtterNotImplementedError):
        raise exc.OtterNotImplementedError

    assert str(exc.OtterNotImplementedError("foo")) == str(
        exc.OtterNotImplementedError()
    )


def test_missing_env_var_error():
    with pytest.raises(exc.MissingEnvVarError):
        raise exc.MissingEnvVarError(["val"])

    e = exc.MissingEnvVarError(["foo"])
    assert e.v == ["foo"]
    assert e.url is None
    assert str(e)[-1] == "\n"

    e = exc.MissingEnvVarError(["foo"], url="fakeurl.com")
    assert e.v == ["foo"]
    assert e.url == "fakeurl.com"
    assert str(e)[-1] == "m"



"""
Unit tests for the otter.Otter class methods
"""

import os
from otter import Otter, Transient
from otter.exceptions import FailedQueryError
from astropy.coordinates import SkyCoord
from astropy.table import Table
import numpy as np
import pandas as pd
import pytest

OTTER_URL = os.environ.get("OTTER_TEST_URL")
OTTER_TEST_PASSWORD = os.environ.get("OTTER_TEST_PASSWORD")


def test_otter_constructor():
    """
    Just make sure everything constructs correctly
    """

    db = Otter(url=OTTER_URL, password=OTTER_TEST_PASSWORD)
    assert isinstance(db, Otter)


def test_get_meta():
    """
    Tests the Otter.get_meta method and make sure it returns as expected
    """

    db = Otter(url=OTTER_URL, password=OTTER_TEST_PASSWORD)

    # first make sure everything is just copied over correctly
    allmeta = db.get_meta()
    true_keys = ["name", "coordinate", "date_reference", "distance", "classification"]

    assert all(k in d for d in allmeta for k in true_keys)

    # now we can try real queries
    metahyz = db.get_meta(names="2018hyz")[0]
    assert isinstance(metahyz, Transient)
    assert metahyz["name/default_name"] == "2018hyz"
    assert metahyz["date_reference"][0]["value"] == "2018-10-14"
    assert metahyz["date_reference"][0]["date_format"] == "iso"
    assert metahyz["classification"]["value"][0]["object_class"] == "TDE"


def test_cone_search():
    """
    Tests the Otter.cone_search method
    """

    db = Otter(url=OTTER_URL, password=OTTER_TEST_PASSWORD)

    # just search around '2018hyz' coordinates to make sure it picks it up
    coord = SkyCoord(151.711964138, 1.69279894089, unit="deg")
    res = db.cone_search(coord)[0]
    assert res["name/default_name"] == "2018hyz"


def test_get_phot():
    """
    Tests the Otter.get_phot method

    We know from the transients.clean_photometry tests that the conversions
    work as expected. So, this will just test that everything comes out as expected.
    """

    db = Otter(url=OTTER_URL, password=OTTER_TEST_PASSWORD)

    true_keys = [
        "name",
        "converted_flux",
        "converted_flux_err",
        "converted_date",
        "converted_wave",
        "converted_freq",
        "converted_flux_unit",
        "converted_date_unit",
        "converted_wave_unit",
        "converted_freq_unit",
        "obs_type",
        "upperlimit",
        "reference",
    ]

    names = ["2018hyz", "2018zr", "ASASSN-14li"]

    # first with returning an astropy table (the default)
    allphot = db.get_phot(names=names)
    assert isinstance(allphot, Table)
    assert all(k in allphot.keys() for k in true_keys)
    assert len(np.unique(allphot["converted_flux_unit"])) == 1
    assert allphot["converted_flux_unit"][0] == "mag(AB)"

    # then with returning a pandas DataFrame
    allphot = db.get_phot(names=names, return_type="pandas")
    assert isinstance(allphot, pd.DataFrame)
    assert all(k in allphot for k in true_keys)
    assert len(np.unique(allphot.converted_flux_unit)) == 1
    assert allphot.converted_flux_unit.iloc[0] == "mag(AB)"

    # then make sure it throws the FailedQueryError
    with pytest.raises(FailedQueryError):
        db.get_phot(names="foo")


def test_query():
    """
    Tests the Otter.query method that basically all of this is based on

    A lot of these have been tested in other unit tests in thie file
    but lets make sure it's complete
    """

    db = Otter(url=OTTER_URL, password=OTTER_TEST_PASSWORD)

    # test min and max z queries
    zgtr1 = db.query(minz=1)
    assert len(zgtr1) >= 2
    true_result = ["Sw J2058+05", "2022cmc", "CXOU J0332"]
    assert all(t["name/default_name"] in true_result for t in zgtr1)

    zless001 = db.query(maxz=0.001)
    result = ["NGC 247", "IGR J17361-4441"]
    assert all(t["name/default_name"] in result for t in zless001)

    # test refs
    # res = db.query(refs="2020MNRAS.tmp.2047S")[0]
    # assert res["name/default_name"] == "2018hyz"

    # test hasphot and hasspec
    assert len(db.query(hasspec=True)) == 0
    assert "ASASSN-20il" not in {t["name/default_name"] for t in db.query(hasphot=True)}



"""
Pydantic Schema Model of our JSON schema
"""

from pydantic import BaseModel, model_validator, field_validator, ValidationError
from typing import Optional, Union, List


class VersionSchema(BaseModel):
    value: Union[str, int] = None
    comment: str = None


class _AliasSchema(BaseModel):
    value: str
    reference: Union[str, List[str]]


class _XrayModelSchema(BaseModel):
    # the following two lines are needed to prevent annoying warnings
    model_config: dict = {}
    model_config["protected_namespaces"] = ()

    # required keywords
    model_name: str
    param_names: List[str]
    param_values: List[Union[float, int, str]]
    param_units: List[Union[str, None]]
    min_energy: Union[float, int, str]
    max_energy: Union[float, int, str]
    energy_units: str

    # optional keywords
    param_value_err_upper: Optional[List[Union[float, int, str]]] = None
    param_value_err_lower: Optional[List[Union[float, int, str]]] = None
    param_upperlimit: Optional[List[Union[float, int, str]]] = None
    param_descriptions: Optional[List[str]] = None
    model_reference: Optional[Union[str, List[str]]] = None


class _ErrDetailSchema(BaseModel):
    # all optional keywords!
    upper: Optional[List[Union[float, int, str]]] = None
    lower: Optional[List[Union[float, int, str]]] = None
    systematic: Optional[List[Union[float, int, str]]] = None
    statistical: Optional[List[Union[float, int, str]]] = None
    iss: Optional[List[Union[float, int, str]]] = None


class NameSchema(BaseModel):
    default_name: str
    alias: list[_AliasSchema]


class CoordinateSchema(BaseModel):
    reference: Union[List[str], str]
    ra: Union[str, float] = None
    dec: Union[str, float] = None
    l: Union[str, float] = None  # noqa: E741
    b: Union[str, float] = None
    lon: Union[str, float] = None
    lat: Union[str, float] = None
    ra_units: str = None
    dec_units: str = None
    l_units: str = None
    b_units: str = None
    lon_units: str = None
    lat_units: str = None
    ra_error: Union[str, float] = None
    dec_error: Union[str, float] = None
    l_error: Union[str, float] = None
    b_error: Union[str, float] = None
    lon_error: Union[str, float] = None
    lat_error: Union[str, float] = None
    epoch: str = None
    frame: str = "J2000"
    coord_type: str = None
    computed: bool = False
    default: bool = False

    @model_validator(mode="after")
    def _has_coordinate(self):
        uses_ra_dec = self.ra is not None and self.dec is not None
        uses_galactic = self.l is not None and self.b is not None
        uses_lon_lat = self.lon is not None and self.lat is not None

        if uses_ra_dec:
            if self.ra_units is None:
                raise ValidationError("ra_units must be provided for RA!")
            if self.dec_units is None:
                raise ValidationError("dec_units must be provided for Dec!")

        elif uses_galactic:
            if self.l_units is None:
                raise ValidationError("l_units must be provided for RA!")
            if self.b_units is None:
                raise ValidationError("b_units must be provided for Dec!")

        elif uses_lon_lat:
            if self.lon_units is None:
                raise ValidationError("lon_units must be provided for RA!")
            if self.lat_units is None:
                raise ValidationError("lat_units must be provided for Dec!")

        else:
            ValidationError("Must have RA/Dec, l/b, and/or lon/lat!")

        return self


class DistanceSchema(BaseModel):
    value: Union[str, float, int]
    unit: str = None
    reference: Union[str, List[str]]
    distance_type: str
    error: Union[str, float, int] = None
    cosmology: str = None
    computed: bool = False
    uuid: str = None
    default: bool = False

    @model_validator(mode="after")
    def _has_units(self):
        if self.distance_type != "redshift" and self.unit is None:
            raise ValidationError("Need units if the distance_type is not redshift!")

        return self


class ClassificationSchema(BaseModel):
    object_class: str
    confidence: float
    reference: Union[str, List[str]]
    default: bool = False
    class_type: str = None


class ClassificationDictSchema(BaseModel):
    spec_classed: Optional[int] = None
    unambiguous: Optional[bool] = None
    value: list[ClassificationSchema]


class ReferenceSchema(BaseModel):
    name: str
    human_readable_name: str


class DateSchema(BaseModel):
    value: Union[str, int, float]
    date_format: str
    date_type: str
    reference: Union[str, List[str]]
    computed: bool = None


class PhotometrySchema(BaseModel):
    reference: Union[List[str], str]
    raw: list[Union[float, int]]
    raw_err: Optional[List[float]] = []
    raw_units: Union[str, List[str]]
    value: Optional[list[Union[float, int]]] = None
    value_err: Optional[list[Union[float, int]]] = None
    value_units: Optional[Union[str, List[str]]] = None
    epoch_zeropoint: Optional[Union[float, str, int]] = None
    epoch_redshift: Optional[Union[float, int]] = None
    filter: Optional[Union[str, List[str]]] = None
    filter_key: Union[str, List[str]]
    obs_type: Union[str, List[str]]
    telescope_area: Optional[Union[float, List[float]]] = None
    date: Union[str, float, List[Union[str, float]]]
    date_format: Union[str, List[str]]
    date_err: Optional[Union[str, float, List[Union[str, float]]]] = None
    ignore: Optional[Union[bool, List[bool]]] = None
    upperlimit: Optional[Union[bool, List[bool]]] = None
    sigma: Optional[Union[str, float, List[Union[str, float]]]] = None
    sky: Optional[Union[str, float, List[Union[str, float]]]] = None
    telescope: Optional[Union[str, List[str]]] = None
    instrument: Optional[Union[str, List[str]]] = None
    phot_type: Optional[Union[str, List[str]]] = None
    exptime: Optional[Union[str, int, float, List[Union[str, int, float]]]] = None
    aperture: Optional[Union[str, int, float, List[Union[str, int, float]]]] = None
    observer: Optional[Union[str, List[str]]] = None
    reducer: Optional[Union[str, List[str]]] = None
    pipeline: Optional[Union[str, List[str]]] = None
    corr_k: Optional[Union[bool, str, List[Union[bool, str]]]] = None
    corr_s: Optional[Union[bool, str, List[Union[bool, str]]]] = None
    corr_av: Optional[Union[bool, str, List[Union[bool, str]]]] = None
    corr_host: Optional[Union[bool, str, List[Union[bool, str]]]] = None
    corr_hostav: Optional[Union[bool, str, List[Union[bool, str]]]] = None
    val_k: Optional[Union[float, int, str, List[Union[float, int, str]]]] = None
    val_s: Optional[Union[float, int, str, List[Union[float, int, str]]]] = None
    val_av: Optional[Union[float, int, str, List[Union[float, int, str]]]] = None
    val_host: Optional[Union[float, int, str, List[Union[float, int, str]]]] = None
    val_hostav: Optional[Union[float, int, str, List[Union[float, int, str]]]] = None
    xray_model: Optional[Union[List[_XrayModelSchema], List[None]]] = None
    raw_err_detail: Optional[_ErrDetailSchema] = None
    value_err_detail: Optional[_ErrDetailSchema] = None

    @field_validator(
        "raw_units",
        "raw_err",
        "filter_key",
        "obs_type",
        "date_format",
        "upperlimit",
        "date",
        "telescope",
    )
    @classmethod
    def ensure_list(cls, v):
        if not isinstance(v, list):
            return [v]
        return v

    @model_validator(mode="after")
    def _ensure_xray_model(self):
        """
        This will eventually ensure the xray_model key is used if obs_type="xray"

        It will be commented out until we get the data setup correctly
        """
        # if self.obs_type == "xray" and self.xray_model is None:
        #     raise ValidationError(
        #         "Need an xray_model for this xray data!"
        #     )

        return self


class FilterSchema(BaseModel):
    filter_key: str
    filter_name: str
    wave_eff: Union[str, float, int] = None
    wave_min: Union[str, float, int] = None
    wave_max: Union[str, float, int] = None
    freq_eff: Union[str, float, int] = None
    freq_min: Union[str, float, int] = None
    freq_max: Union[str, float, int] = None
    zp: Union[str, float, int] = None
    wave_units: Union[str, float, int] = None
    freq_units: Union[str, float, int] = None
    zp_units: Union[str, float, int] = None
    zp_system: Union[str, float, int] = None


class HostSchema(BaseModel):
    reference: Union[str, List[str]]
    host_ra: Optional[Union[str, float]] = None
    host_dec: Optional[Union[str, float]] = None
    host_ra_units: Optional[str] = None
    host_dec_units: Optional[str] = None
    host_z: Optional[Union[str, int, float]] = None
    host_type: Optional[str] = None
    host_name: Optional[str] = None

    @model_validator(mode="after")
    def _has_coordinate_or_name(self):
        has_coordinate = self.host_ra is not None and self.host_dec is not None
        has_name = self.host_name is not None

        # if it has the RA/Dec keys, make sure it also has ra_unit, dec_unit keys
        if has_coordinate:
            if self.host_ra_units is None:
                raise ValidationError("Need RA unit if coordinates are provided!")
            if self.host_dec_units is None:
                raise ValidationError("Need Dec unit if coordinates are provided!")

        # we need either the coordinate or name to identify this object
        # Both are okay too (more info is always better)
        if not has_coordinate and not has_name:
            raise ValidationError(
                "Need to provide a Host name and/or host coordinates!"
            )

        # Make sure that if one of RA/Dec is given then both are given
        if (self.host_ra is None and self.host_dec is not None) or (
            self.host_ra is not None and self.host_dec is None
        ):
            raise ValidationError(
                "Please provide RA AND Dec, not just one or the other!"
            )

        return self


class OtterSchema(BaseModel):
    schema_version: Optional[VersionSchema] = None
    name: NameSchema
    coordinate: list[CoordinateSchema]
    distance: Optional[list[DistanceSchema]] = None
    classification: Optional[ClassificationDictSchema] = None
    reference_alias: list[ReferenceSchema]
    date_reference: Optional[list[DateSchema]] = None
    photometry: Optional[list[PhotometrySchema]] = None
    filter_alias: Optional[list[FilterSchema]] = None
    host: Optional[list[HostSchema]] = None

    @model_validator(mode="after")
    def _verify_filter_alias(self):
        if self.photometry is not None and self.filter_alias is None:
            raise ValidationError("filter_alias is needed if photometry is given!")



"""
Host object that stores information on the Transient Host and provides utility methods
for pulling in data corresponding to that host
"""

from __future__ import annotations

from urllib.request import urlopen
import json

import numpy as np
from astropy.coordinates import SkyCoord
from astropy import units as u

from .data_finder import DataFinder
from ..exceptions import OtterLimitationError

import logging

logger = logging.getLogger(__name__)


class Host(DataFinder):
    def __init__(
        self,
        host_ra: str | float,
        host_dec: str | float,
        host_ra_units: str | u.Unit,
        host_dec_units: str | u.Unit,
        host_name: str = None,
        host_redshift: float = None,
        redshift_type: str = None,
        reference: list[str] = None,
        transient_name: str = None,
        **kwargs,
    ) -> None:
        """
        Object to store host information and query public data sources of host galaxies

        Subclass of the data scraper class to allow for these queries to happen

        Args:
            host_ra (str|float) : The RA of the host to be passed to an astropy SkyCoord
            host_dec (str|float) : The declination of the host to be passed to an
                                   astropy SkyCoord
            host_ra_units (str|astropy.units.Unit) : units of the RA, to be passed to
                                                     the unit keyword of SkyCoord
            host_dec_units (str|astropy.units.Unit) : units of the declination, to be
                                                      passed to the unit keyword of
                                                      SkyCoord
            host_name (str) : The name of the host galaxy
            host_redshift (float) : The redshift of the host galaxy
            redshift_type (str) : Either "phot" or "spec", tells you if the redshift
                                  is a phot-z or spec-z.
            reference (list[str]) : a list of bibcodes that found this to be the host
            transient_name (str) : the name of the transient associated with this host
            kwargs : Just here so we can pass **Transient['host'] into this constructor
                     and any extraneous properties will be ignored.
        """
        self.coord = SkyCoord(host_ra, host_dec, unit=(host_ra_units, host_dec_units))
        self.name = host_name
        self.z = host_redshift
        self.redshift = host_redshift  # just here for ease of use
        self.bibcodes = reference
        self.transient_name = transient_name
        self.redshift_type = redshift_type

    def pcc(self, transient_coord: SkyCoord, mag: float = None):
        """
        Compute the Probability of Chance Coincindence as described in
        Bloom et al. (2002) "Offset Distribution of Gamma-Ray Bursts.

        This computes the probability that this galaxy is by chance nearby to the
        transient on the sky. Or, in simpler terms this essentially computes the
        probability that we are wrong about this being the transient host. So, a
        smaller probability is better!

        Note: This probability was initially defined for GRB afterglows, which tend to
        be redder transients (supernova too). So, be cautious when using this algorithm
        for TDEs!

        Args:
            transient_coord (astropy.coordinates.SkyCoord) : The coordinates of the
                                                             transient object.
            mag (float) : An r-band magnitude to compute from. Default is None which
                          will prompt us to check SDSS for one within 10".
        Returns:
            A float probability in the range [0,1]
        """

        # first get the separation r, in arcseconds
        r = self.coord.separation(transient_coord).arcsec

        # next get the host r magnitude
        if mag is None:
            res = self.query_vizier(radius=10 * u.arcsec)

            if len(res) == 0:
                raise OtterLimitationError(
                    "No magnitude found in SDSS! Please provide a magnitude via the \
                    `mag` keyword to make this calculation!"
                )

            sdss = [k for k in res.keys() if "sdss" in k]
            use = max(sdss, key=lambda k: int(k.split("sdss")[-1]))
            print(f"Using the r magnitude from the {use} table")
            mag = res[use]["rmag"][0]

        # then compute the probability
        sigma_prefactor = 1 / (3600**2 * 0.334 * np.log(10))
        sigma_pow = 0.334 * (mag - 22.963) + 4.320
        sigma = sigma_prefactor * 10**sigma_pow

        eta = np.pi * r**2 * sigma

        prob = 1 - np.exp(-eta)

        return prob

    ###################################################################################
    ######### METHODS FOR FINDING HOSTS  ###########################
    ###################################################################################
    @staticmethod
    def query_blast(tns_name: str) -> dict:
        """
        Query the BLAST host galaxy service

        Args:
            tns_name (str) : The TNS target name to grab the data from BLAST for

        Returns:
            best match BLAST ID, host ra, host dec, host redshift. Redshift will be
            spectroscopic if available, otherwise photometric.
        """

        # clean up the input name a little
        if tns_name[:2] == "AT":
            tns_name = tns_name.replace("AT", "")

        failed_query_res = None

        # do the query
        blast_base_url = "https://blast.scimma.org"
        blast_query_url = f"{blast_base_url}/api/transient/?name={tns_name}&format=json"
        with urlopen(blast_query_url) as response:
            if response.status != 200:
                logger.warn(f"BLAST query failed with response code {response.status}!")
                return failed_query_res
            else:
                blast_data = json.loads(response.read())
        if len(blast_data) == 0:
            logger.warn("BLAST query returned no results!")
            return failed_query_res

        blast_data = blast_data[0]
        if "host" not in blast_data:
            logger.warn("BLAST query found the object but it has no host associated!")
            return failed_query_res

        blast_host = blast_data["host"]

        if blast_host["redshift"] is not None:
            # prefer spec-z over phot-z
            z = blast_host["redshift"]
            z_type = "spec"
        else:
            # well I guess we need to use phot-z
            z = blast_host["photometric_redshift"]
            z_type = "phot"

        refs = [
            "2021ApJ...908..170G",  # GHOST citation
            "2024arXiv241017322J",  # BLAST citation
        ]

        return Host(
            host_ra=blast_host["ra_deg"],
            host_dec=blast_host["dec_deg"],
            host_ra_units="deg",
            host_dec_units="deg",
            host_name=blast_host["id"],
            host_redshift=z,
            redshift_type=z_type,
            reference=refs,
            transient_name=tns_name,
        )



"""
Class for a transient,
basically just inherits the dict properties with some overwriting
"""

from __future__ import annotations
import warnings
from copy import deepcopy
import re
from collections.abc import MutableMapping
from typing_extensions import Self
import logging

import numpy as np
import pandas as pd

import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord

from ..exceptions import (
    FailedQueryError,
    IOError,
    OtterLimitationError,
    TransientMergeError,
)
from ..util import XRAY_AREAS, _KNOWN_CLASS_ROOTS, _DuplicateFilter
from .host import Host

np.seterr(divide="ignore")
logger = logging.getLogger(__name__)


class Transient(MutableMapping):
    def __init__(self, d={}, name=None):
        """
        Overwrite the dictionary init

        Args:
            d (dict): A transient dictionary
            name (str): The default name of the transient, default is None and it will
                        be inferred from the input dictionary.
        """
        self.data = d

        if "reference_alias" in self:
            self.srcmap = {
                ref["name"]: ref["human_readable_name"]
                for ref in self["reference_alias"]
            }
            self.srcmap["TNS"] = "TNS"
        else:
            self.srcmap = {}

        if "name" in self:
            if "default_name" in self["name"]:
                self.default_name = self["name"]["default_name"]
            else:
                raise AttributeError("Missing the default name!!")
        elif name is not None:
            self.default_name = name
        else:
            self.default_name = "Missing Default Name"

        # Make it so all coordinates are astropy skycoords

    def __getitem__(self, keys):
        """
        Override getitem to recursively access Transient elements
        """

        if isinstance(keys, (list, tuple)):
            return Transient({key: (self[key] if key in self else []) for key in keys})
        elif isinstance(keys, str) and "/" in keys:  # this is for a path
            s = "']['".join(keys.split("/"))
            s = "['" + s
            s += "']"
            return eval(f"self{s}")
        elif (
            isinstance(keys, int)
            or keys.isdigit()
            or (keys[0] == "-" and keys[1:].isdigit())
        ):
            # this is for indexing a sublist
            return self[int(keys)]
        else:
            return self.data[keys]

    def __setitem__(self, key, value):
        """
        Override set item to work with the '/' syntax
        """

        if isinstance(key, str) and "/" in key:  # this is for a path
            s = "']['".join(key.split("/"))
            s = "['" + s
            s += "']"
            exec(f"self{s} = value")
        else:
            self.data[key] = value

    def __delitem__(self, keys):
        if "/" in keys:
            raise OtterLimitationError(
                "For security, we can not delete with the / syntax!"
            )
        else:
            del self.data[keys]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __repr__(self, html=False):
        if not html:
            return f"Transient(\n\tName: {self.default_name},\n\tKeys: {self.keys()}\n)"
        else:
            html = ""

            coord = self.get_skycoord()

            # add the ra and dec
            # These are required so no need to check if they are there
            html += f"""
            <tr>
            <td style="text-align:left">RA [hrs]:</td>
            <td style="text-align:left">{coord.ra}
            </tr>
            <tr>
            <td style="text-align:left">DEC [deg]:</td>
            <td style="text-align:left">{coord.dec}
            </tr>
            """

            if "date_reference" in self:
                discovery = self.getDiscoveryDate().to_value("datetime")
                if discovery is not None:
                    # add the discovery date
                    html += f"""
                    <tr>
                    <td style="text-align:left">Discovery Date [MJD]:</td>
                    <td style="text-align:left">{discovery}
                    </tr>
                    """

            if "distance" in self:
                # add the redshift
                html += f"""
                <tr>
                <td style="text-align:left">Redshift:</td>
                <td style="text-align:left">{self['distance']['redshift'][0]['value']}
                </tr>
                """

            if "reference_alias" in self:
                srcs = ""
                for bibcode, src in self.srcmap.items():
                    srcs += f"<a href='https://ui.adsabs.harvard.edu/abs/{bibcode}'"
                    srcs += f"target='_blank'>{src}</a><br>"

                html += f"""
                <tr>
                <td style="text-align:left">Sources:</td>
                <td style="text-align:left">{srcs}
                </tr>
                """

            return html

    def keys(self):
        return self.data.keys()

    def __add__(self, other, strict_merge=True):
        """
        Merge this transient object with another transient object

        Args:
            other [Transient]: A Transient object to merge with
            strict_merge [bool]: If True it won't let you merge objects that
                                 intuitively shouldn't be merged (ie. different
                                 transient events).
        """

        # first check that this object is within a good distance of the other object
        if (
            strict_merge
            and self.get_skycoord().separation(other.get_skycoord()) > 10 * u.arcsec
        ):
            raise TransientMergeError(
                "These two transients are not within 10 arcseconds!"
                + " They probably do not belong together! If they do"
                + " You can set strict_merge=False to override the check"
            )

        # create set of the allowed keywords
        allowed_keywords = {
            "name",
            "date_reference",
            "coordinate",
            "distance",
            "filter_alias",
            "schema_version",
            "photometry",
            "classification",
            "host",
        }

        merge_subkeys_map = {
            "name": None,
            "date_reference": ["value", "date_format", "date_type"],
            "coordinate": None,  # may need to update this if we run into problems
            "distance": ["value", "distance_type", "unit"],
            "filter_alias": None,
            "schema_version": None,
            "photometry": None,
            "classification": None,
            "host": [
                "host_ra",
                "host_dec",
                "host_ra_units",
                "host_dec_units",
                "host_name",
            ],
        }

        groupby_key_for_default_map = {
            "name": None,
            "date_reference": "date_type",
            "coordinate": "coordinate_type",
            "distance": "distance_type",
            "filter_alias": None,
            "schema_version": None,
            "photometry": None,
            "classification": None,
            "host": None,
        }

        # create a blank dictionary since we don't want to overwrite this object
        out = {}

        # find the keys that are
        merge_keys = list(
            self.keys() & other.keys()
        )  # in both t1 and t2 so we need to merge these keys
        only_in_t1 = list(self.keys() - other.keys())  # only in t1
        only_in_t2 = list(other.keys() - self.keys())  # only in t2

        # now let's handle the merge keys
        for key in merge_keys:
            # reference_alias is special
            # we ALWAYS should combine these two
            if key == "reference_alias":
                out[key] = self[key]
                if self[key] != other[key]:
                    # only add t2 values if they aren't already in it
                    bibcodes = {ref["name"] for ref in self[key]}
                    for val in other[key]:
                        if val["name"] not in bibcodes:
                            out[key].append(val)
                continue

            # we can skip this merge process and just add the values from t1
            # if they are equal. We should still add the new reference though!
            if self[key] == other[key]:
                # set the value
                # we don't need to worry about references because this will
                # only be true if the reference is also equal!
                out[key] = deepcopy(self[key])
                continue

            # There are some special keys that we are expecting
            if key in allowed_keywords:
                Transient._merge_arbitrary(
                    key,
                    self,
                    other,
                    out,
                    merge_subkeys=merge_subkeys_map[key],
                    groupby_key=groupby_key_for_default_map[key],
                )
            else:
                # this is an unexpected key!
                if strict_merge:
                    # since this is a strict merge we don't want unexpected data!
                    raise TransientMergeError(f"{key} was not expected! Can not merge!")
                else:
                    # Throw a warning and only keep the old stuff
                    logger.warning(
                        f"{key} was not expected! Only keeping the old information!"
                    )
                    out[key] = deepcopy(self[key])

        # and now combining out with the stuff only in t1 and t2
        out = out | dict(self[only_in_t1]) | dict(other[only_in_t2])

        # now return out as a Transient Object
        return Transient(out)

    def get_meta(self, keys=None) -> Self:
        """
        Get the metadata (no photometry or spectra)

        This essentially just wraps on __getitem__ but with some checks

        Args:
            keys (list[str]) : list of keys to get the metadata for from the transient

        Returns:
            A Transient object of just the meta data
        """
        if keys is None:
            keys = list(self.keys())

            # note: using the remove method is safe here because dict keys are unique
            if "photometry" in keys:
                keys.remove("photometry")
            if "spectra" in keys:
                keys.remove("spectra")
        else:
            # run some checks
            if "photometry" in keys:
                logger.warning("Not returing the photometry!")
                _ = keys.pop("photometry")
            if "spectra" in keys:
                logger.warning("Not returning the spectra!")
                _ = keys.pop("spectra")

            curr_keys = self.keys()
            for key in keys:
                if key not in curr_keys:
                    keys.remove(key)
                    logger.warning(
                        f"Not returning {key} because it is not in this transient!"
                    )

        return self[keys]

    def get_skycoord(self, coord_format="icrs") -> SkyCoord:
        """
        Convert the coordinates to an astropy SkyCoord

        Args:
            coord_format (str): Astropy coordinate format to convert the SkyCoord to
                                defaults to icrs.

        Returns:
            Astropy.coordinates.SkyCoord of the default coordinate for the transient
        """

        # now we can generate the SkyCoord
        f = "df['coordinate_type'] == 'equatorial'"
        coord_dict = self._get_default("coordinate", filt=f)
        coordin = self._reformat_coordinate(coord_dict)
        coord = SkyCoord(**coordin).transform_to(coord_format)

        return coord

    def get_discovery_date(self) -> Time:
        """
        Get the default discovery date for this Transient

        Returns:
            astropy.time.Time of the default discovery date
        """
        key = "date_reference"
        try:
            date = self._get_default(key, filt='df["date_type"] == "discovery"')
        except KeyError:
            return None

        if date is None:
            return date

        if "date_format" in date:
            f = date["date_format"]
        else:
            f = "mjd"

        return Time(str(date["value"]).strip(), format=f)

    def get_redshift(self) -> float:
        """
        Get the default redshift of this Transient

        Returns:
            Float value of the default redshift
        """
        f = "df['distance_type']=='redshift'"
        default = self._get_default("distance", filt=f)
        if default is None:
            return default
        else:
            return default["value"]

    def get_classification(self) -> tuple(str, float, list):
        """
        Get the default classification of this Transient.
        This normally corresponds to the highest confidence classification that we have
        stored for the transient.

        Returns:
            The default object class as a string, the confidence level in that class,
            and a list of the bibcodes corresponding to that classification. Or, None
            if there is no classification.
        """
        default = self._get_default("classification/value")
        if default is None:
            return default
        return default.object_class, default.confidence, default.reference

    def get_host(self, max_hosts=3, search=False, **kwargs) -> list[Host]:
        """
        Gets the default host information of this Transient. This returns an otter.Host
        object. If search=True, it will also check the BLAST host association database
        for the best match and return it as well. Note that if search is True then
        this has the potential to return max_hosts + 1, if BLAST also returns a result.
        The BLAST result will always be the last value in the returned list.

        Args:
            max_hosts [int] : The maximum number of hosts to return, default is 3
            **kwargs : keyword arguments to be passed to getGHOST

        Returns:
            A list of otter.Host objects. This is useful becuase the Host objects have
            useful methods for querying public catalogs for data of the host.
        """
        # first try to get the host information from our local database
        host = []
        if "host" in self:
            max_hosts = min([max_hosts, len(self["host"])])
            for h in self["host"][:max_hosts]:
                host.append(Host(transient_name=self.default_name, **dict(h)))

        # then try BLAST
        if search:
            logger.warning(
                "Trying to find a host with BLAST/astro-ghost. Note\
                 that this won't work for older targets! See https://blast.scimma.org"
            )

            # default_name should always be the TNS name if we have one
            print(self.default_name)
            blast_host = Host.query_blast(self.default_name)
            print(blast_host)
            if blast_host is not None:
                host.append(blast_host)

        return host

    def _get_default(self, key, filt=None):
        """
        Get the default of key

        Args:
            key [str]: key in self to look for the default of
            filt [str]: a valid pandas dataframe filter to index a pandas dataframe
                        called df.
        """
        if key not in self:
            raise KeyError(f"This transient does not have {key} associated with it!")

        df = pd.DataFrame(self[key])
        if len(df) == 0:
            raise KeyError(f"This transient does not have {key} associated with it!")

        if filt is not None:
            df = df[eval(filt)]  # apply the filters

        if "default" in df:
            # first try to get the default
            df_filtered = df[df.default == True]
            if len(df_filtered) == 0:
                df_filtered = df
        else:
            df_filtered = df

        if len(df_filtered) == 0:
            return None

        return df_filtered.iloc[0]

    def _reformat_coordinate(self, item):
        """
        Reformat the coordinate information in item
        """
        coordin = None
        if "ra" in item and "dec" in item:
            # this is an equatorial coordinate
            coordin = {
                "ra": item["ra"],
                "dec": item["dec"],
                "unit": (item["ra_units"], item["dec_units"]),
            }
        elif "l" in item and "b" in item:
            coordin = {
                "l": item["l"],
                "b": item["b"],
                "unit": (item["l_units"], item["b_units"]),
                "frame": "galactic",
            }

        return coordin

    def clean_photometry(
        self,
        flux_unit: u.Unit = "mag(AB)",
        date_unit: u.Unit = "MJD",
        freq_unit: u.Unit = "GHz",
        wave_unit: u.Unit = "nm",
        obs_type: str = None,
    ) -> pd.DataFrame:
        """
        Ensure the photometry associated with this transient is all in the same
        units/system/etc

        Args:
            flux_unit (astropy.unit.Unit): The astropy unit or string representation of
                                           an astropy unit to convert and return the
                                           flux as.
            date_unit (str): Valid astropy date format string.
            freq_unit (astropy.unit.Unit): The astropy unit or string representation of
                                           an astropy unit to convert and return the
                                           frequency as.
            wave_unit (astropy.unit.Unit): The astropy unit or string representation of
                                           an astropy unit to convert and return the
                                           wavelength as.
            obs_type (str): "radio", "xray", or "uvoir". If provided, it only returns
                            data taken within that range of wavelengths/frequencies.
                            Default is None which will return all of the data.

        Returns:
            A pandas DataFrame of the cleaned up photometry in the requested units
        """
        warn_filt = _DuplicateFilter()
        logger.addFilter(warn_filt)

        # these imports need to be here for some reason
        # otherwise the code breaks
        from synphot.units import VEGAMAG, convert_flux
        from synphot.spectrum import SourceSpectrum

        # variable so this warning only displays a single time each time this
        # function is called
        source_map_warning = True

        # turn the photometry key into a pandas dataframe
        if "photometry" not in self:
            raise FailedQueryError("No photometry for this object!")

        dfs = []
        for item in self["photometry"]:
            max_len = 0
            for key, val in item.items():
                if isinstance(val, list) and key != "reference":
                    max_len = max(max_len, len(val))

            for key, val in item.items():
                if not isinstance(val, list) or (
                    isinstance(val, list) and len(val) != max_len
                ):
                    item[key] = [val] * max_len

            df = pd.DataFrame(item)
            dfs.append(df)

        if len(dfs) == 0:
            raise FailedQueryError("No photometry for this object!")
        c = pd.concat(dfs)

        # extract the filter information and substitute in any missing columns
        # because of how we handle this later, we just need to make sure the effective
        # wavelengths are never nan
        def fill_wave(row):
            if "wave_eff" not in row or (
                pd.isna(row.wave_eff) and not pd.isna(row.freq_eff)
            ):
                freq_eff = row.freq_eff * u.Unit(row.freq_units)
                wave_eff = freq_eff.to(u.Unit(wave_unit), equivalencies=u.spectral())
                return wave_eff.value, wave_unit
            elif not pd.isna(row.wave_eff):
                return row.wave_eff, row.wave_units
            else:
                raise ValueError("Missing frequency or wavelength information!")

        filters = pd.DataFrame(self["filter_alias"])
        res = filters.apply(fill_wave, axis=1)
        filters["wave_eff"], filters["wave_units"] = zip(*res)
        # merge the photometry with the filter information
        df = c.merge(filters, on="filter_key")

        # drop irrelevant obs_types before continuing
        if obs_type is not None:
            valid_obs_types = {"radio", "uvoir", "xray"}
            if obs_type not in valid_obs_types:
                raise IOError("Please provide a valid obs_type")
            df = df[df.obs_type == obs_type]

        # add some mockup columns if they don't exist
        if "value" not in df:
            df["value"] = np.nan
            df["value_err"] = np.nan
            df["value_units"] = "NaN"

        # fix some bad units that are old and no longer recognized by astropy
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            df.raw_units = df.raw_units.str.replace("ergs", "erg")
            df.raw_units = ["mag(AB)" if uu == "AB" else uu for uu in df.raw_units]
            df.value_units = df.value_units.str.replace("ergs", "erg")
            df.value_units = ["mag(AB)" if uu == "AB" else uu for uu in df.value_units]

        # merge the raw and value keywords based on the requested flux_units
        # first take everything that just has `raw` and not `value`
        df_raw_only = df[df.value.isna()]
        remaining = df[df.value.notna()]
        if len(remaining) == 0:
            df_raw = df_raw_only
            df_value = []  # this tricks the code later
        else:
            # then take the remaining rows and figure out if we want the raw or value
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                flux_unit_astropy = u.Unit(flux_unit)

                val_unit_filt = np.array(
                    [
                        u.Unit(uu).is_equivalent(flux_unit_astropy)
                        for uu in remaining.value_units
                    ]
                )

            df_value = remaining[val_unit_filt]
            df_raw_and_value = remaining[~val_unit_filt]

            # then merge the raw dataframes
            df_raw = pd.concat([df_raw_only, df_raw_and_value], axis=0)

        # then add columns to these dataframes to convert stuff later
        df_raw = df_raw.assign(
            _flux=df_raw["raw"].values,
            _flux_units=df_raw["raw_units"].values,
            _flux_err=(
                df_raw["raw_err"].values
                if "raw_err" in df_raw
                else [np.nan] * len(df_raw)
            ),
        )

        if len(df_value) == 0:
            df = df_raw
        else:
            df_value = df_value.assign(
                _flux=df_value["value"].values,
                _flux_units=df_value["value_units"].values,
                _flux_err=(
                    df_value["value_err"].values
                    if "value_err" in df_value
                    else [np.nan] * len(df_value)
                ),
            )

            # then merge df_value and df_raw back into one df
            df = pd.concat([df_raw, df_value], axis=0)

        # then, for the rest of the code to work, set the "by" variables to _flux
        by = "_flux"

        # skip rows where 'by' is nan
        df = df[df[by].notna()]

        # remove rows where the flux is less than zero since this is nonphysical
        # See Mummery et al. (2023) Section 5.2 for why we need to do this when using
        # ZTF data:
        # "Because the origin of the negative late-time flux is currently un-
        # known (and under investigation), we have not attempted to correct
        # the TDE lightcurves for this systematic effect. "
        df = df[df[by].astype(float) > 0]

        # filter out anything that has _flux_units == "ct" because we can't convert that
        try:
            # this is a test case to see if we can convert ct -> flux_unit
            convert_flux(
                [1 * u.nm, 2 * u.nm], 1 * u.ct, u.Unit(flux_unit), area=1 * u.m**2
            )
        except u.UnitsError:
            bad_units = df[df._flux_units == "ct"]
            if len(bad_units) > 0:
                logger.warning(
                    f"""Removing {len(bad_units)} photometry points from
                    {self.default_name} because we can't convert them from ct ->
                    {flux_unit}"""
                )
            df = df[df._flux_units != "ct"]

        # convert the ads bibcodes to a string of human readable sources here
        def mappedrefs(row):
            if isinstance(row.reference, list):
                return "<br>".join([self.srcmap[bibcode] for bibcode in row.reference])
            else:
                return self.srcmap[row.reference]

        try:
            df["human_readable_refs"] = df.apply(mappedrefs, axis=1)
        except Exception as exc:
            if source_map_warning:
                source_map_warning = False
                logger.warning(f"Unable to apply the source mapping because {exc}")

            df["human_readable_refs"] = df.reference

        # Figure out what columns are good to groupby in the photometry
        outdata = []

        if "telescope" in df:
            tele = True
            to_grp_by = ["obs_type", by + "_units", "telescope"]
        else:
            tele = False
            to_grp_by = ["obs_type", by + "_units"]

        # Do the conversion based on what we decided to group by
        for groupedby, data in df.groupby(to_grp_by, dropna=False):
            if tele:
                obstype, unit, telescope = groupedby
            else:
                obstype, unit = groupedby
                telescope = None

            # get the photometry in the right type
            unit = data[by + "_units"].unique()
            if len(unit) > 1:
                raise OtterLimitationError(
                    "Can not apply multiple units for different obs_types"
                )

            unit = unit[0]
            isvegamag = "vega" in unit.lower()
            try:
                if isvegamag:
                    astropy_units = VEGAMAG
                elif unit == "AB":
                    # In astropy "AB" is a magnitude SYSTEM not unit and while
                    # u.Unit("AB") will succeed without error, it will not produce
                    # the expected result!
                    # We can assume here that this unit really means astropy's "mag(AB)"
                    astropy_units = u.Unit("mag(AB)")
                else:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        astropy_units = u.Unit(unit)

            except ValueError:
                # this means there is something likely slightly off in the input unit
                # string. Let's try to fix it!
                # here are some common mistakes
                unit = unit.replace("ergs", "erg")
                unit = unit.replace("AB", "mag(AB)")

                astropy_units = u.Unit(unit)

            except ValueError:
                raise ValueError(
                    "Could not coerce your string into astropy unit format!"
                )

            # get the flux data and find the type
            indata = np.array(data[by].astype(float))
            err_key = by + "_err"
            if err_key in data:
                indata_err = np.array(data[by + "_err"].astype(float))
            else:
                indata_err = np.zeros(len(data))

            # convert to an astropy quantity
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                q = indata * u.Unit(astropy_units)
                q_err = indata_err * u.Unit(
                    astropy_units
                )  # assume error and values have the same unit

            # get and save the effective wavelength
            # because of cleaning we did to the filter dataframe above wave_eff
            # should NEVER be nan!
            if np.any(pd.isna(data["wave_eff"])):
                raise ValueError("Flushing out the effective wavelength array failed!")

            zz = zip(data["wave_eff"], data["wave_units"])
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                wave_eff = u.Quantity([vv * u.Unit(uu) for vv, uu in zz], wave_unit)
                freq_eff = wave_eff.to(freq_unit, equivalencies=u.spectral())

            data["converted_wave"] = wave_eff.value
            data["converted_wave_unit"] = wave_unit
            data["converted_freq"] = freq_eff.value
            data["converted_freq_unit"] = freq_unit

            # convert using synphot
            # stuff has to be done slightly differently for xray than for the others
            if obstype == "xray":
                if telescope is not None:
                    try:
                        area = XRAY_AREAS[telescope.lower()]
                    except KeyError:
                        raise OtterLimitationError(
                            "Did not find an area corresponding to "
                            + "this telescope, please add to util!"
                        )
                else:
                    raise OtterLimitationError(
                        "Can not convert x-ray data without a telescope"
                    )

                # we also need to make this wave_min and wave_max
                # instead of just the effective wavelength like for radio and uvoir
                zz = zip(data["wave_min"], data["wave_max"], data["wave_units"])
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore")
                    wave_eff = u.Quantity(
                        [np.array([m, M]) * u.Unit(uu) for m, M, uu in zz],
                        u.Unit(wave_unit),
                    )

            else:
                area = None

            if obstype == "xray" or isvegamag:
                # we unfortunately have to loop over the points here because
                # syncphot does not work with a 2D array of min max wavelengths
                # for converting counts to other flux units. It also can't convert
                # vega mags with a wavelength array because it interprets that as the
                # wavelengths corresponding to the SourceSpectrum.from_vega()

                flux, flux_err = [], []
                for wave, xray_point, xray_point_err in zip(wave_eff, q, q_err):
                    with warnings.catch_warnings():
                        warnings.filterwarnings("ignore")
                        f_val = convert_flux(
                            wave,
                            xray_point,
                            u.Unit(flux_unit),
                            vegaspec=SourceSpectrum.from_vega(),
                            area=area,
                        ).value

                    # approximate the uncertainty as dX = dY/Y * X
                    f_err = np.multiply(
                        f_val, np.divide(xray_point_err.value, xray_point.value)
                    )

                    # then we take the average of the minimum and maximum values
                    # computed by syncphot
                    flux.append(np.mean(f_val))
                    flux_err.append(np.mean(f_err))

            else:
                # this will be faster and cover most cases
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore")
                    flux = convert_flux(wave_eff, q, u.Unit(flux_unit)).value

                # since the error propagation is different between logarithmic units
                # and linear units, unfortunately
                if isinstance(u.Unit(flux_unit), u.LogUnit):
                    # approximate the uncertainty as dX = dY/Y * |ln(10)/2.5|
                    prefactor = np.abs(np.log(10) / 2.5)  # this is basically 1
                else:
                    # approximate the uncertainty as dX = dY/Y * X
                    prefactor = flux

                flux_err = np.multiply(prefactor, np.divide(q_err.value, q.value))

            flux = np.array(flux) * u.Unit(flux_unit)
            flux_err = np.array(flux_err) * u.Unit(flux_unit)

            data["converted_flux"] = flux.value
            data["converted_flux_err"] = flux_err.value
            outdata.append(data)

        if len(outdata) == 0:
            raise FailedQueryError()
        outdata = pd.concat(outdata)

        # copy over the flux units
        outdata["converted_flux_unit"] = flux_unit

        # make sure all the datetimes are in the same format here too!!
        times = [
            Time(d, format=f).to_value(date_unit.lower())
            for d, f in zip(outdata.date, outdata.date_format.str.lower())
        ]
        outdata["converted_date"] = times
        outdata["converted_date_unit"] = date_unit

        # compute the upperlimit value based on a 3 sigma detection
        # this is just for rows where we don't already know if it is an upperlimit
        if isinstance(u.Unit(flux_unit), u.LogUnit):
            # this uses the following formula (which is surprising because it means
            # magnitude upperlimits are independent of the actual measurement!)
            # sigma_m > (1/3) * (ln(10)/2.5)
            def is_upperlimit(row):
                if "upperlimit" in row and pd.isna(row.upperlimit):
                    return row.converted_flux_err > np.log(10) / (3 * 2.5)
                else:
                    return row.upperlimit
        else:

            def is_upperlimit(row):
                if "upperlimit" in row and pd.isna(row.upperlimit):
                    return row.converted_flux < 3 * row.converted_flux_err
                else:
                    return row.upperlimit

        outdata["upperlimit"] = outdata.apply(is_upperlimit, axis=1)

        logger.removeFilter(warn_filt)
        return outdata

    def _merge_names(t1, t2, out):  # noqa: N805
        """
        Private method to merge the name data in t1 and t2 and put it in out
        """
        key = "name"
        out[key] = {}

        # first deal with the default_name key
        # we are gonna need to use some regex magic to choose a preferred default_name
        if t1[key]["default_name"] == t2[key]["default_name"]:
            out[key]["default_name"] = t1[key]["default_name"]
        else:
            # we need to decide which default_name is better
            # it should be the one that matches the TNS style
            # let's use regex
            n1 = t1[key]["default_name"]
            n2 = t2[key]["default_name"]

            # write some discriminating regex expressions
            # exp1: starts with a number, this is preferred because it is TNS style
            exp1 = "^[0-9]"
            # exp2: starts with any character, also preferred because it is TNS style
            exp2 = ".$"
            # exp3: checks if first four characters are a number, like a year :),
            # this is pretty strict though
            exp3 = "^[0-9]{3}"
            # exp4: # checks if it starts with AT like TNS names
            exp4 = "^AT"

            # combine all the regex expressions, this makes it easier to add more later
            exps = [exp1, exp2, exp3, exp4]

            # score each default_name based on this
            score1 = 0
            score2 = 0
            for e in exps:
                re1 = re.findall(e, n1)
                re2 = re.findall(e, n2)
                if re1:
                    score1 += 1
                if re2:
                    score2 += 1

            # assign a default_name based on the score
            if score1 > score2:
                out[key]["default_name"] = t1[key]["default_name"]
            elif score2 > score1:
                out[key]["default_name"] = t2[key]["default_name"]
            else:
                logger.warning(
                    "Names have the same score! Just using the existing default_name"
                )
                out[key]["default_name"] = t1[key]["default_name"]

        # now deal with aliases
        # create a reference mapping for each
        t1map = {}
        for val in t1[key]["alias"]:
            ref = val["reference"]
            if isinstance(ref, str):
                t1map[val["value"]] = [ref] if isinstance(ref, str) else list(ref)
            else:
                t1map[val["value"]] = [ref] if isinstance(ref, str) else list(ref)

        t2map = {}
        for val in t2[key]["alias"]:
            ref = val["reference"]
            if isinstance(ref, str):
                t2map[val["value"]] = [ref] if isinstance(ref, str) else list(ref)
            else:
                t2map[val["value"]] = [ref] if isinstance(ref, str) else list(ref)

        # figure out which ones we need to be careful with references in
        inboth = list(
            t1map.keys() & t2map.keys()
        )  # in both so we'll have to merge the reference key
        int1 = list(t1map.keys() - t2map.keys())  # only in t1
        int2 = list(t2map.keys() - t1map.keys())  # only in t2

        # add ones that are not in both first, these are easy
        line1 = [{"value": k, "reference": t1map[k]} for k in int1]
        line2 = [{"value": k, "reference": t2map[k]} for k in int2]
        bothlines = [{"value": k, "reference": t1map[k] + t2map[k]} for k in inboth]
        out[key]["alias"] = line2 + line1 + bothlines

    def _merge_filter_alias(t1, t2, out):  # noqa: N805
        """
        Combine the filter alias lists across the transient objects
        """

        key = "filter_alias"

        out[key] = deepcopy(t1[key])
        keys1 = {filt["filter_key"] for filt in t1[key]}
        for filt in t2[key]:
            if filt["filter_key"] not in keys1:
                out[key].append(filt)

    def _merge_schema_version(t1, t2, out):  # noqa: N805
        """
        Just keep whichever schema version is greater
        """
        key = "schema_version/value"
        if "comment" not in t1["schema_version"]:
            t1["schema_version/comment"] = ""

        if "comment" not in t2["schema_version"]:
            t2["schema_version/comment"] = ""

        if key in t1 and key in t2 and int(t1[key]) > int(t2[key]):
            out["schema_version"] = deepcopy(t1["schema_version"])
        else:
            out["schema_version"] = deepcopy(t2["schema_version"])

        out["schema_version"]["comment"] = (
            t1["schema_version/comment"] + ";" + t2["schema_version/comment"]
        )

    def _merge_photometry(t1, t2, out):  # noqa: N805
        """
        Combine photometry sources
        """

        key = "photometry"

        out[key] = deepcopy(t1[key])
        refs = []  # np.array([d["reference"] for d in out[key]])
        # merge_dups = lambda val: np.sum(val) if np.any(val.isna()) else val.iloc[0]
        for val in out[key]:
            if isinstance(val, list):
                refs += val
            elif isinstance(val, np.ndarray):
                refs += list(val)
            else:
                refs.append(val)

        for val in t2[key]:
            # first check if t2's reference is in out
            if val["reference"] not in refs:
                # it's not here so we can just append the new photometry!
                out[key].append(val)
            else:
                # we need to merge it with other photometry
                i1 = np.where(val["reference"] == refs)[0][0]
                df1 = pd.DataFrame(out[key][i1])
                df2 = pd.DataFrame(val)

                # only substitute in values that are nan in df1 or new
                # the combined keys of the two
                mergeon = list(set(df1.keys()) & set(df2.keys()))
                df = df1.merge(df2, on=mergeon, how="outer")
                # convert to a dictionary
                newdict = df.reset_index().to_dict(orient="list")
                del newdict["index"]

                newdict["reference"] = newdict["reference"][0]

                out[key][i1] = newdict  # replace the dictionary at i1 with the new dict

    def _merge_class(t1, t2, out):  # noqa: N805
        """
        Combine the classification attribute
        """
        key = "classification"
        subkey = "value"
        out[key] = deepcopy(t1[key])
        classes = np.array([item["object_class"] for item in out[key][subkey]])

        for item in t2[key][subkey]:
            if item["object_class"] in classes:
                i = np.where(item["object_class"] == classes)[0][0]
                if int(item["confidence"]) > int(out[key][subkey][i]["confidence"]):
                    out[key][subkey][i]["confidence"] = item[
                        "confidence"
                    ]  # we are now more confident

                if not isinstance(out[key][subkey][i]["reference"], list):
                    out[key][subkey][i]["reference"] = [
                        out[key][subkey][i]["reference"]
                    ]

                if not isinstance(item["reference"], list):
                    item["reference"] = [item["reference"]]

                newdata = list(
                    np.unique(out[key][subkey][i]["reference"] + item["reference"])
                )
                out[key][subkey][i]["reference"] = newdata

            else:
                out[key][subkey].append(item)

        # now that we have all of them we need to figure out which one is the default
        maxconf = max(out[key][subkey], key=lambda d: d["confidence"])
        for item in out[key][subkey]:
            if item == maxconf:
                item["default"] = True
            else:
                item["default"] = False

        # then rederive the classification flags
        out = Transient._derive_classification_flags(out)

    @classmethod
    def _derive_classification_flags(cls, out):
        """
        Derive the classification flags based on the confidence flags. This will find
        - spec_classed
        - unambiguous

        See the paper for a detailed description of how this algorithm makes its
        choices
        """

        if "classification" not in out or "value" not in out["classification"]:
            # this means that the transient doesn't have any classifications
            # just return itself without any changes
            return out

        # get the confidences of all of the classifications of this transient
        confs = np.array(
            [item["confidence"] for item in out["classification"]["value"]]
        ).astype(float)

        all_class_roots = np.array(
            [
                _fuzzy_class_root(item["object_class"])
                for item in out["classification"]["value"]
            ]
        )

        if np.any(confs >= 3):
            unambiguous = len(np.unique(all_class_roots)) == 1
            if np.any(confs == 3) or np.any(confs == 3.3):
                # this is a "gold spectrum"
                spec_classed = 3
            elif np.any(confs == 3.2):
                # this is a silver spectrum
                spec_classed = 2
            elif np.any(confs == 3.1):
                # this is a bronze spectrum
                spec_classed = 1
            else:
                raise ValueError("Not prepared for this confidence flag!")

        elif np.any(confs == 2):
            # these always have spec_classed = True, by definition
            # They also have unambiguous = False by definition because they don't
            # have a peer reviewed citation for their classification
            spec_classed = 1
            unambiguous = False

        elif np.any(confs == 1):
            spec_classed = 0  # by definition
            unambiguous = len(np.unique(all_class_roots)) == 1

        else:
            spec_classed = 0
            unambiguous = False

        # finally, set these keys in the classification dict
        out["classification"]["spec_classed"] = spec_classed
        out["classification"]["unambiguous"] = unambiguous

        return out

    @staticmethod
    def _merge_arbitrary(key, t1, t2, out, merge_subkeys=None, groupby_key=None):
        """
        Merge two arbitrary datasets inside the json file using pandas

        The datasets in t1 and t2 in "key" must be able to be forced into
        a NxM pandas dataframe!
        """

        if key == "name":
            t1._merge_names(t2, out)
        elif key == "filter_alias":
            t1._merge_filter_alias(t2, out)
        elif key == "schema_version":
            t1._merge_schema_version(t2, out)
        elif key == "photometry":
            t1._merge_photometry(t2, out)
        elif key == "classification":
            t1._merge_class(t2, out)
        else:
            # this is where we can standardize some of the merging
            df1 = pd.DataFrame(t1[key])
            df2 = pd.DataFrame(t2[key])

            merged_with_dups = pd.concat([df1, df2]).reset_index(drop=True)

            # have to get the indexes to drop using a string rep of the df
            # this is cause we have lists in some cells
            # We also need to deal with merging the lists of references across rows
            # that we deem to be duplicates. This solution to do this quickly is from
            # https://stackoverflow.com/questions/36271413/ \
            # pandas-merge-nearly-duplicate-rows-based-on-column-value
            if merge_subkeys is None:
                merge_subkeys = merged_with_dups.columns.tolist()
                merge_subkeys.remove("reference")
            else:
                for k in merge_subkeys:
                    if k not in merged_with_dups:
                        merge_subkeys.remove(k)

            merged = (
                merged_with_dups.astype(str)
                .groupby(merge_subkeys)["reference"]
                .apply(lambda x: x.sum())
                .reset_index()
            )

            # then we have to turn the merged reference strings into a string list
            merged["reference"] = merged.reference.str.replace("][", ",")

            # then eval the string of a list to get back an actual list of sources
            merged["reference"] = merged.reference.apply(
                lambda v: np.unique(eval(v)).tolist()
            )

            # decide on default values
            if groupby_key is None:
                iterate_through = [(0, merged)]
            else:
                iterate_through = merged.groupby(groupby_key)

            # we will make whichever value has more references the default
            outdict = []
            for data_type, df in iterate_through:
                lengths = df.reference.map(len)
                max_idx_arr = np.argmax(lengths)

                if isinstance(max_idx_arr, np.int64):
                    max_idx = max_idx_arr
                elif len(max_idx_arr) == 0:
                    raise ValueError("Something went wrong with deciding the default")
                else:
                    max_idx = max_idx_arr[0]  # arbitrarily choose the first

                defaults = np.full(len(df), False, dtype=bool)
                defaults[max_idx] = True

                df["default"] = defaults
                outdict.append(df)
            outdict = pd.concat(outdict)

            # from https://stackoverflow.com/questions/52504972/ \
            # converting-a-pandas-df-to-json-without-nan
            outdict = outdict.replace("nan", np.nan)
            outdict_cleaned = [{**x[i]} for i, x in outdict.stack().groupby(level=0)]

            out[key] = outdict_cleaned


def _fuzzy_class_root(s):
    """
    Extract the fuzzy classification root name from the string s
    """
    s = s.upper()
    # first split the class s using regex
    for root in _KNOWN_CLASS_ROOTS:
        if s.startswith(root):
            remaining = s[len(root) :]
            if remaining and root == "SN":
                # we want to be able to distinguish between SN Ia and SN II
                # we will use SN Ia to indicate thoes and SN to indicate CCSN
                if "IA" in remaining or "1A" in remaining:
                    return "SN Ia"
            return root
    return s



"""
Test the otter Host class
"""

from otter import Host
from astropy.coordinates import SkyCoord
from urllib.error import URLError


def test_constructor():
    """
    Test the Host constructor
    """
    host_ra = 192.06249999999997
    host_dec = 17.77401111111111
    host_name = "SDSS J124815.23+174626.4"
    h = Host(
        host_ra=host_ra,
        host_dec=host_dec,
        host_name=host_name,
        host_ra_units="deg",
        host_dec_units="deg",
        transient_name="ASASSN-14li",
        reference=["foobar"],
        host_redshift=0.1,
    )

    assert isinstance(h.coord, SkyCoord), "Not constructing a SkyCoord"
    assert h.coord.ra.value == host_ra, "Incorrect RA"
    assert h.coord.dec.value == host_dec, "Incorrect Dec"
    assert h.name == host_name
    assert h.z == 0.1
    assert h.redshift == 0.1
    assert h.bibcodes == ["foobar"]
    assert h.transient_name == "ASASSN-14li"


def test_pcc():
    """
    Test the PCC score testing
    """
    host_ra = 192.06249999999997
    host_dec = 17.77401111111111
    host_name = "SDSS J124815.23+174626.4"
    h = Host(
        host_ra=host_ra,
        host_dec=host_dec,
        host_name=host_name,
        host_ra_units="deg",
        host_dec_units="deg",
        transient_name="ASASSN-14li",
        reference=["foobar"],
        host_redshift=0.1,
    )

    # get the testing path
    coord = SkyCoord(192.06343875, 17.77402083, unit="deg")
    pcc = h.pcc(coord)

    assert pcc < 0.001, "pcc seems suspiciously high for ASASSN-14li as a test case..."

    pcc2 = h.pcc(SkyCoord(0, 0, unit="deg"))
    assert pcc2 > 0.9, "pcc seems too small for this input coordinate"


def test_query_blast():
    """
    Test the query of BLAST
    """

    try:
        host = Host.query_blast(tns_name="AT2024advm")
    except URLError:
        return  # this isn't our fault, probably

    assert host.transient_name == "2024advm", "name incorrect!"



from __future__ import annotations

# get the version
from ._version import __version__


# explicitly set the package variable to ensure relative import work
__package__ = "otter"

# import important stuff
from .io.otter import Otter
from .io.transient import Transient
from .io.host import Host
from .io.data_finder import DataFinder
from .plotter.otter_plotter import OtterPlotter
from .plotter.plotter import plot_light_curve, plot_sed, quick_view, query_quick_view
from . import util
from . import schema
from . import exceptions



"""
Custom exceptions for otter
"""

from __future__ import annotations


class FailedQueryError(ValueError):
    """
    Exception thrown when the users query does not return any results.
    """

    def __str__(self):
        txt = "You're query/search did not return any results! "
        txt += "Try again with different parameters!"
        return txt


class IOError(ValueError):
    """
    Exception thrown when the input or output argument/value is not the correct type.
    """

    pass


class OtterLimitationError(Exception):
    """
    Exception thrown when the user requests something that is currently not supported
    by the API.
    """

    def __init__(self, msg):
        self.msg = "Current Limitation Found: " + msg

    def __str__(self):
        return self.msg


class TransientMergeError(Exception):
    """
    Exception thrown when the Transient objects can not be combined as expected.
    """

    pass


class OtterNotImplementedError(Exception):
    """
    Exception thrown by methods that are not yet implemented
    """

    def __str__(self):
        return "Not yet implemented! Please consider contributing on our GitHub!"


class MissingEnvVarError(Exception):
    """
    Thrown when the querying code expects an Environment variable and doesn't get one
    """

    def __init__(self, v, url=None):
        self.v = v
        self.url = url

    def __str__(self):
        msg = "You are missing the following environment variables:\n"
        for vv in self.v:
            msg += f"\t- {vv}\n"
        msg += "Please make an account and/or set these variables!\n"
        if self.url is not None:
            msg += f"See: {self.url}"

        return msg



"""
Host object that stores information on the Transient DataFinder and provides utility
methods for pulling in data corresponding to that host
"""

from __future__ import annotations
import os
import csv
import io
import re
import time
import math
from urllib.request import urlopen
import requests

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy.table import Table
from astropy.io.votable import parse_single_table
from astropy.io import ascii

import numpy as np
import pandas as pd
import logging

from fundamentals.stats import rolling_window_sigma_clip
from operator import itemgetter

from ..util import VIZIER_LARGE_CATALOGS
from ..exceptions import MissingEnvVarError

logger = logging.getLogger(__name__)


class DataFinder(object):
    def __init__(
        self,
        ra: str | float,
        dec: str | float,
        ra_units: str | u.Unit,
        dec_units: str | u.Unit,
        name: str = None,
        redshift: float = None,
        reference: list[str] = None,
        **kwargs,
    ) -> None:
        """
        Object to store DataFinder info to query public data sources of host galaxies

        Args:
            ra (str|float) : The RA of the host to be passed to an astropy SkyCoord
            dec (str|float) : The declination of the host to be passed to an
                                   astropy SkyCoord
            ra_units (str|astropy.units.Unit) : units of the RA, to be passed to
                                                     the unit keyword of SkyCoord
            dec_units (str|astropy.units.Unit) : units of the declination, to be
                                                      passed to the unit keyword of
                                                      SkyCoord
            name (str) : The name of the host galaxy
            redshift (float) : The redshift of the host galaxy
            reference (list[str]) : a list of bibcodes that found this to be the host
            kwargs : Just here so we can pass **Transient['host'] into this constructor
                     and any extraneous properties will be ignored.
        """
        self.coord = SkyCoord(ra, dec, unit=(ra_units, dec_units))
        self.name = name
        self.z = redshift
        self.redshift = redshift  # just here for ease of use
        self.bibcodes = reference

    def __repr__(self) -> str:
        """
        String representation of the DataFinder for printing
        """

        if self.name is None:
            print_name = "No Name DataFinder"
        else:
            print_name = self.name

        return f"{print_name} @ (RA, Dec)=({self.coord.ra},{self.coord.dec})"

    def __iter__(self) -> dict:
        """
        Provides an iterator for the properties of this DataFinder. Yields (key, value)
        """
        out = dict(
            host_ra=self.coord.ra.value,
            host_dec=self.coord.dec.value,
            host_ra_units="deg",
            host_dec_units="deg",
        )

        if self.name is not None:
            out["host_name"] = self.name

        if self.z is not None:
            out["host_redshift"] = self.z

        if self.bibcodes is not None:
            out["reference"] = self.bibcodes

        for k, v in out.items():
            yield (k, v)

    ###################################################################################
    ################### CONVENIENCE METHODS FOR QUERYING HOST METADATA ################
    ###################################################################################

    @staticmethod
    def _wrap_astroquery(module, *args, **kwargs):
        """
        Private convenience method that just standardizes how we call the query_region
        method in astroquery
        """
        return module.query_region(*args, **kwargs)

    def query_simbad(self, radius="5 arcsec", **kwargs):
        """
        Query SIMBAD through astroquery to provide any other "meta" information on this
        host that may not be stored in the OTTER

        Args:
            radius (str|astropy.quantity.Quantity) : search radius for astroquery
            **kwargs : any other arguments for astroquery.vizier.Vizier.query_region

        Returns:
            astropy Table of the simbad results.
        """
        from astroquery.simbad import Simbad

        return DataFinder._wrap_astroquery(Simbad, self.coord, radius=radius, **kwargs)

    def query_vizier(self, radius="5 arcsec", **kwargs):
        """
        Query the ViZier catalog for TIME-AVERAGED data from their major/large catalogs.

        ViZier Catalogs Queried:
           - 2MASS-PSC
           - 2MASX
           - AC2000.2
           - AKARI
           - ALLWISE
           - ASCC-2.5
           - B/DENIS
           - CMC14
           - Gaia-DR1
           - GALEX
           - GLIMPSE
           - GSC-ACT
           - GSC1.2
           - GSC2.2
           - GSC2.3
           - HIP
           - HIP2
           - IRAS
           - NOMAD1
           - NVSS
           - PanSTARRS-DR1
           - PGC
           - Planck-DR1
           - PPMX
           - PPMXL
           - SDSS-DR12
           - SDSS-DR7
           - SDSS-DR9
           - Tycho-2
           - UCAC2
           - UCAC3
           - UCAC4
           - UKIDSS
           - USNO-A2
           - USNO-B1
           - WISE

        Args:
            radius (str|astropy.quantity.Quantity) : search radius for astroquery
            **kwargs : any other arguments for astroquery.vizier.Vizier.query_region

        Returns:
            astropy TableList of the time-averaged photometry associated with this host.
        """
        from astroquery.vizier import Vizier

        return DataFinder._wrap_astroquery(
            Vizier, self.coord, radius=radius, catalog=VIZIER_LARGE_CATALOGS
        )

    ###################################################################################
    ######### CONVENIENCE METHODS FOR QUERYING HOST TIME SERIES PHOTOMETRY  ###########
    ###################################################################################

    def query_atlas(
        self, days_ago: int = 365, disc_date: float = None, clip_sigma: float = 2.0
    ) -> pd.DataFrame:
        """
        Query ATLAS forced photometry for photometry for this host

        Args:
            days_ago (int) : Number of days before the transients discovery date
                             (or today if no disc_date is given) to get ATLAS
                             forced photometry for.
            disc_date (float) : The discovery date of the transient in MJD.
            clip_sigma (float) : amount to sigma clip the ATLAS data by

        Return:
            pandas DataFrame of the ATLAS forced photometry for this host
        """
        base_url = "https://fallingstar-data.com/forcedphot"

        token = os.environ.get("ATLAS_API_TOKEN", None)
        if token is None:
            logger.warn(
                "Getting your token from ATLAS. Please add ATLAS_API_TOKEN to your \
                environment variables to avoid this!"
            )

            uname = os.environ.get("ATLAS_UNAME", default=None)
            pword = os.environ.get("ATLAS_PWORD", default=None)

            if uname is None and pword is None:
                raise MissingEnvVarError(["ATLAS_UNAME", "ATLAS_PWORD"], base_url)
            elif uname is None and pword is not None:
                raise MissingEnvVarError(["ATLAS_UNAME"], base_url)
            elif uname is not None and pword is None:
                raise MissingEnvVarError(["ATLAS_PWORD"], base_url)

            resp = requests.post(
                url=f"{base_url}/api-token-auth/",
                data={"username": uname, "password": pword},
            )

            token = resp.json()["token"]

        headers = {"Authorization": f"Token {token}", "Accept": "application/json"}

        # compute the query start
        if disc_date is None:
            t_queryend = Time.now().mjd
            logger.warn(
                "Since no transient name is given we are using today \
                as the query end!"
            )
        else:
            t_queryend = Time(disc_date, format="mjd").mjd

        t_querystart = t_queryend - days_ago

        # submit the query to the ATLAS forced photometry server
        task_url = None
        while not task_url:
            with requests.Session() as s:
                resp = s.post(
                    f"{base_url}/queue/",
                    headers=headers,
                    data={
                        "ra": self.coord.ra.value,
                        "dec": self.coord.ra.value,
                        "send_email": False,
                        "mjd_min": t_querystart,
                        "mjd_max": t_queryend,
                        "use_reduced": False,
                    },
                )
                if resp.status_code == 201:  # success
                    task_url = resp.json()["url"]
                    logger.info(f"The task URL is {task_url}")
                elif resp.status_code == 429:  # throttled
                    message = resp.json()["detail"]
                    logger.info(f"{resp.status_code} {message}")
                    t_sec = re.findall(r"available in (\d+) seconds", message)
                    t_min = re.findall(r"available in (\d+) minutes", message)
                    if t_sec:
                        waittime = int(t_sec[0])
                    elif t_min:
                        waittime = int(t_min[0]) * 60
                    else:
                        waittime = 10
                    logger.info(f"Waiting {waittime} seconds")
                    time.sleep(waittime)
                else:
                    raise Exception(f"ERROR {resp.status_code}\n{resp.text}")

        # Now wait for the result
        result_url = None
        taskstarted_printed = False
        while not result_url:
            with requests.Session() as s:
                resp = s.get(task_url, headers=headers)

                if resp.status_code == 200:  # HTTP OK
                    if resp.json()["finishtimestamp"]:
                        result_url = resp.json()["result_url"]
                        logger.info(
                            f"Task is complete with results available at {result_url}"
                        )
                    elif resp.json()["starttimestamp"]:
                        if not taskstarted_printed:
                            print(
                                f"Task is running (started at\
                                {resp.json()['starttimestamp']})"
                            )
                            taskstarted_printed = True
                        time.sleep(2)
                    else:
                        # print(f"Waiting for job to start (queued at {timestamp})")
                        time.sleep(4)
                else:
                    raise Exception(f"ERROR {resp.status_code}\n{resp.text}")

        # get and clean up the result
        with requests.Session() as s:
            textdata = s.get(result_url, headers=headers).text

        atlas_phot = DataFinder._atlas_stack(textdata, clipping_sigma=clip_sigma)

        return pd.DataFrame(atlas_phot)

    def query_ptf(self, radius: str | u.Quantity = "5 arcsec", **kwargs) -> Table:
        """
        Query the palomer transient facility's light curve catalog for this host

        Args:
            radius (str|astropy.quantity.Quantity) : search radius
            **kwargs : other optional arguments for astroquery's query_region

        Returns:
            An astropy Table of the resulting light curve
        """
        from astroquery.ipac.irsa import Irsa

        ptf_lc_catalog = "ptf_lightcurves"
        return DataFinder._wrap_astroquery(
            Irsa, self.coord, radius=radius, catalog=ptf_lc_catalog
        )

    def query_ztf(self, radius: float = 5):
        """
        Query ZTF photometry/forced photometry for photometry for this host

        Args:
            radius (float) : The search radius in arcseconds

        Returns:
            An astropy table of the time series data from the cone search in ZTF
        """

        base_url = "https://irsa.ipac.caltech.edu/cgi-bin/ZTF/nph_light_curves?"

        ra, dec = self.coord.ra.value, self.coord.dec.value
        search_radius_arcseconds = radius  # in arcseconds
        search_radius_degree = search_radius_arcseconds / 3600

        query_url = f"{base_url}POS=CIRCLE%20{ra}%20{dec}%20{search_radius_degree}"

        resp = urlopen(query_url)

        votab = parse_single_table(io.BytesIO(resp.read()))

        return Table(votab.array)

    def query_asassn(self, radius: float = 5.0, nthreads: int = 2) -> pd.DataFrame:
        """
        Query ASASSN photometry/forced photometry for photometry for this host

        Args:
            radius (float) : search radius in arcseconds
            nthreads (int) : number of threads to utilize during download, default is 2

        Returns:
            A pandas dataframe with the ASASSN lightcurve for this object
        """
        from pyasassn.client import SkyPatrolClient

        client = SkyPatrolClient()
        light_curve = client.cone_search(
            self.coord.ra.value,
            self.coord.dec.value,
            radius=radius,
            units="arcsec",
            download=True,
            threads=nthreads,
        )
        return light_curve.data

    def query_wise(
        self,
        radius: float = 5,
        datadir: str = "ipac/",
        overwrite: bool = False,
        verbose=False,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Query NEOWISE for their multiepoch photometry

        The method used to query wise here was taken from this github repo:
        https://github.com/HC-Hwang/wise_light_curves/tree/master
        and you should cite this other paper that the authors of this code developed
        it for: https://ui.adsabs.harvard.edu/abs/2020MNRAS.493.2271H/abstract

        This will download the ipac data files to the "datadir" argument. by default,
        these will go into os.getcwd()/ipac

        Args:
            radius (float) : The cone search radius in arcseconds
            overwrite (bool) : Overwrite the existing datasets downloaded from wise
            **kwargs : Other optional arguments for the astroquery query_region
        Returns:
            An astropy Table of the multiepoch wise data for this host
        """
        # from https://www.cambridge.org/core/journals/
        # publications-of-the-astronomical-society-of-australia/article/
        # recalibrating-the-widefield-infrared-survey-explorer-wise-w4-filter/
        # B238BFFE19A533A2D2638FE88CCC2E89
        band_vals = {"w1": 3.4, "w2": 4.6, "w3": 12, "w4": 22}  # in um

        ra, dec = self.coord.ra.value, self.coord.dec.value

        fbasename = f"wise_{self.name}"
        allwise_name = f"{fbasename}_allwise.ipac"
        neowise_name = f"{fbasename}_neowise.ipac"

        if not os.path.exists(datadir):
            os.makedirs(datadir)

        self._download_single_data(
            name=fbasename,
            ra=ra,
            dec=dec,
            root_path=datadir,
            radius=radius,
            overwrite=overwrite,
        )

        allwise = ascii.read(f"ipac/{allwise_name}", format="ipac")
        neowise = ascii.read(f"ipac/{neowise_name}", format="ipac")

        allwise, neowise = self._only_good_data(allwise, neowise, verbose=verbose)
        if verbose and (allwise is None or neowise is None):
            print(f"Limited good infrared data for {self.name}, skipping!")

        mjd, mag, mag_err, filts = self._make_full_lightcurve_multibands(
            allwise, neowise, bands=["w1", "w2", "w3", "w4"]
        )

        df = pd.DataFrame(
            dict(
                name=[self.name] * len(mjd),
                date_mjd=mjd,
                filter=filts,
                filter_eff=[band_vals[f] for f in filts],
                filter_eff_unit=["um"] * len(mjd),
                flux=mag,
                flux_err=mag_err,
                flux_unit=["mag(AB)"] * len(mjd),
                upperlimit=[False] * len(mjd),
            )
        )

        # clean up the wise data by filtering out negative flux
        wise = df[df.flux > 0].reset_index(drop=True)
        return wise

    def query_alma(self, radius: float = 5, **kwargs) -> Table:
        """
        Query ALMA to see if there are observations of this host.

        NOTE: Since this is radio/mm data, it is unlikely that the output table will
        simply have fluxes in it. Instead you will need to use the access_url column
        to download and reduce this data.

        Args:
            radius (float) : The cone search radius in arcseconds
            **kwargs : Other optional arguments for the astroquery query_region
        Returns:
            An astropy Table of the multiepoch wise data for this host
        """

        logger.warn(
            "This method may not work if you are using a conda environment!\
            This is a known issue in setuptools that is not resolved!"
        )

        from astroquery.alma import Alma

        res = DataFinder._wrap_astroquery(
            Alma, self.coord, radius=5 * u.arcsec, **kwargs
        )
        return res

    def query_first(
        self, radius: u.Quantity = 5 * u.arcmin, get_image: bool = False, **kwargs
    ) -> list:
        """
        Query the FIRST radio survey and return an astropy table of the flux density

        This queries Table 2 from Ofek & Frail (2011); 2011ApJ...737...45O

        Args:
            radius (u.Quantity) : An astropy Quantity with the image height/width
            get_image (bool) : If True, download and return a list of the associated
                               images too.
            **kwargs : any other arguments to pass to the astroquery.image_cutouts
                       get_images method

        Returns:
            Astropy table of the flux densities. If get_image is True, it also returns
            a list of FIRST radio survey images
        """
        from astroquery.vizier import Vizier

        res = DataFinder._wrap_astroquery(
            Vizier, self.coord, radius=radius, catalog="J/ApJ/737/45/table2"
        )

        if get_image:
            from astroquery.image_cutouts.first import First

            res_img = First.get_images(self.coord, image_size=radius, **kwargs)
            return res, res_img

        return res

    def query_nvss(self, radius: u.Quantity = 5 * u.arcsec, **kwargs) -> Table:
        """
        Query the NRAO VLA Sky Survey (NVSS) and return a table list of the
        result

        This queries Table 1 from Ofek & Frail (2011); 2011ApJ...737...45O

        Args:
            radius (u.Quantity) : An astropy Quantity with the radius
            **kwargs : Any other arguments to pass to query_region
        """
        from astroquery.vizier import Vizier

        res = DataFinder._wrap_astroquery(
            Vizier, self.coord, radius=radius, catalog="J/ApJ/737/45/table1"
        )
        return res

    def query_heasarc(self, radius: u.Quantity = 5 * u.arcsec, **kwargs) -> Table:
        """
        Query Heasarc by the argument "heasarc_key" for the ra/dec associated with this
        DataLoader object.

        Args:
            radius (u.Quantity) : An astropy Quantity with the radius
            heasarc_table (str) : String with name of heasarc table to query. Default is
                                'xray' which queries the heasarc master x-ray catalog,
                                'radio' will query the heasarc master radio catalog. See
                                https://heasarc.gsfc.nasa.gov/cgi-bin/W3Browse/w3catindex.pl
                                for a complete list.
            **kwargs : Any other arguments to pass to query_region

        Returns:
            Astropy table of the rows in `heasarc_table` that match self.coord.
        """
        from astroquery.heasarc import Heasarc

        res = DataFinder._wrap_astroquery(Heasarc, self.coord, radius=radius, **kwargs)

        return res

    ###################################################################################
    ######### CONVENIENCE METHODS FOR QUERYING HOST SPECTR  ###########################
    ###################################################################################

    def query_sparcl(
        self, radius: u.Quantity = 5 * u.arcsec, include: str | list = "DEFAULT"
    ) -> Table:
        """
        Query the NOIRLab DataLabs Sparcl database for spectra for this host

        Args:
            radius (Quantity) : search radius as an Astropy.unit.Quantity
            include [list|str] : list or string of columns to include in the result. See
                                 the sparcl client documentation for more info. The
                                 default returns specid, ra, dec, sparcl_id, flux,
                                 wavelength, and the spectroscopic surveyu (_dr)

        Returns:
            astropy Table of the results, one row per spectrum
        """

        from sparcl.client import SparclClient
        from dl import queryClient as qc  # noqa: N813

        client = SparclClient()

        # first do a cone search on sparcl.main
        ra, dec = self.coord.ra.value, self.coord.dec.value
        radius_deg = radius.to(u.deg).value

        adql = f"""
        SELECT *
        FROM sparcl.main
        WHERE 't'=Q3C_RADIAL_QUERY(ra,dec,{ra},{dec},{radius_deg})
        """
        cone_search_res = qc.query(adql=adql, fmt="pandas")

        # then retrieve all of the spectra corresponding to those sparcl_ids
        sparcl_ids = cone_search_res.sparcl_id.tolist()
        res = client.retrieve(uuid_list=sparcl_ids, include=include)
        all_spec = pd.concat([pd.DataFrame([record]) for record in res.records])
        return Table.from_pandas(all_spec)

    ###################################################################################
    ######### PRIVATE HELPER METHODS FOR THE QUERYING #################################
    ###################################################################################
    @staticmethod
    def _atlas_stack(filecontent, clipping_sigma, log=logger):
        """
        Function adapted from David Young's :func:`plotter.plot_single_result`
        https://github.com/thespacedoctor/plot-results-from-atlas-force-photometry-service/blob/main/plot_atlas_fp.py

        And again adapted from https://github.com/SAGUARO-MMA/kne-cand-vetting/blob/master/kne_cand_vetting/survey_phot.py
        """
        epochs = DataFinder._atlas_read_and_sigma_clip_data(
            filecontent, log=log, clipping_sigma=clipping_sigma
        )

        # c = cyan, o = arange
        magnitudes = {
            "c": {"mjds": [], "mags": [], "magErrs": [], "lim5sig": []},
            "o": {"mjds": [], "mags": [], "magErrs": [], "lim5sig": []},
            "I": {"mjds": [], "mags": [], "magErrs": [], "lim5sig": []},
        }

        # SPLIT BY FILTER
        for epoch in epochs:
            if epoch["F"] in ["c", "o", "I"]:
                magnitudes[epoch["F"]]["mjds"].append(epoch["MJD"])
                magnitudes[epoch["F"]]["mags"].append(epoch["uJy"])
                magnitudes[epoch["F"]]["magErrs"].append(epoch["duJy"])
                magnitudes[epoch["F"]]["lim5sig"].append(epoch["mag5sig"])

        # STACK PHOTOMETRY IF REQUIRED
        stacked_magnitudes = DataFinder._stack_photometry(magnitudes, binningdays=1)

        return stacked_magnitudes

    @staticmethod
    def _atlas_read_and_sigma_clip_data(filecontent, log, clipping_sigma=2.2):
        """
        Function adapted from David Young's :func:`plotter.read_and_sigma_clip_data`
        https://github.com/thespacedoctor/plot-results-from-atlas-force-photometry-service/blob/main/plot_atlas_fp.py

        And again adapted from
        https://github.com/SAGUARO-MMA/kne-cand-vetting/blob/master/kne_cand_vetting/survey_phot.py

        *clean up rouge data from the files by performing some basic clipping*
        **Key Arguments:**
        - `fpFile` -- path to single force photometry file
        - `clippingSigma` -- the level at which to clip flux data
        **Return:**
        - `epochs` -- sigma clipped and cleaned epoch data
        """

        # CLEAN UP FILE FOR EASIER READING
        fpdata = (
            filecontent.replace("###", "")
            .replace(" ", ",")
            .replace(",,", ",")
            .replace(",,", ",")
            .replace(",,", ",")
            .replace(",,", ",")
            .splitlines()
        )

        # PARSE DATA WITH SOME FIXED CLIPPING
        oepochs = []
        cepochs = []
        csvreader = csv.DictReader(
            fpdata, dialect="excel", delimiter=",", quotechar='"'
        )

        for row in csvreader:
            for k, v in row.items():
                try:
                    row[k] = float(v)
                except Exception:
                    pass
            # REMOVE VERY HIGH ERROR DATA POINTS, POOR CHI SQUARED, OR POOR EPOCHS
            if row["duJy"] > 4000 or row["chi/N"] > 100 or row["mag5sig"] < 17.0:
                continue
            if row["F"] == "c":
                cepochs.append(row)
            if row["F"] == "o":
                oepochs.append(row)

        # SORT BY MJD
        cepochs = sorted(cepochs, key=itemgetter("MJD"), reverse=False)
        oepochs = sorted(oepochs, key=itemgetter("MJD"), reverse=False)

        # SIGMA-CLIP THE DATA WITH A ROLLING WINDOW
        cdataflux = []
        cdataflux[:] = [row["uJy"] for row in cepochs]
        odataflux = []
        odataflux[:] = [row["uJy"] for row in oepochs]

        masklist = []
        for flux in [cdataflux, odataflux]:
            fullmask = rolling_window_sigma_clip(
                log=log, array=flux, clippingSigma=clipping_sigma, windowSize=11
            )
            masklist.append(fullmask)

        try:
            cepochs = [e for e, m in zip(cepochs, masklist[0]) if m == False]
        except Exception:
            cepochs = []

        try:
            oepochs = [e for e, m in zip(oepochs, masklist[1]) if m == False]
        except Exception:
            oepochs = []

        logger.info("Completed the ``read_and_sigma_clip_data`` function")
        # Returns ordered dictionary of all parameters
        return cepochs + oepochs

    @staticmethod
    def _stack_photometry(magnitudes, binningdays=1.0):
        """
        Function adapted from David Young's :func:`plotter.stack_photometry`
        https://github.com/thespacedoctor/plot-results-from-atlas-force-photometry-service/blob/main/plot_atlas_fp.py

        And again adapted from
        https://github.com/SAGUARO-MMA/kne-cand-vetting/blob/master/kne_cand_vetting/survey_phot.py

        *stack the photometry for the given temporal range*
        **Key Arguments:**
            - `magnitudes` -- dictionary of photometry divided into filter sets
            - `binningDays` -- the binning to use (in days)
        **Return:**
            - `summedMagnitudes` -- the stacked photometry
        """

        # IF WE WANT TO 'STACK' THE PHOTOMETRY
        summed_magnitudes = {
            "c": {"mjds": [], "mags": [], "magErrs": [], "n": [], "lim5sig": []},
            "o": {"mjds": [], "mags": [], "magErrs": [], "n": [], "lim5sig": []},
            "I": {"mjds": [], "mags": [], "magErrs": [], "n": [], "lim5sig": []},
        }

        # MAGNITUDES/FLUXES ARE DIVIDED IN UNIQUE FILTER SETS - SO ITERATE OVER
        # FILTERS
        alldata = []
        for fil, data in list(magnitudes.items()):
            # WE'RE GOING TO CREATE FURTHER SUBSETS FOR EACH UNQIUE MJD
            # (FLOORED TO AN INTEGER)
            # MAG VARIABLE == FLUX (JUST TO CONFUSE YOU)
            distinctmjds = {}
            for mjd, flx, err, lim in zip(
                data["mjds"], data["mags"], data["magErrs"], data["lim5sig"]
            ):
                # DICT KEY IS THE UNIQUE INTEGER MJD
                key = str(int(math.floor(mjd / float(binningdays))))
                # FIRST DATA POINT OF THE NIGHTS? CREATE NEW DATA SET
                if key not in distinctmjds:
                    distinctmjds[key] = {
                        "mjds": [mjd],
                        "mags": [flx],
                        "magErrs": [err],
                        "lim5sig": [lim],
                    }
                # OR NOT THE FIRST? APPEND TO ALREADY CREATED LIST
                else:
                    distinctmjds[key]["mjds"].append(mjd)
                    distinctmjds[key]["mags"].append(flx)
                    distinctmjds[key]["magErrs"].append(err)
                    distinctmjds[key]["lim5sig"].append(lim)

            # ALL DATA NOW IN MJD SUBSETS. SO FOR EACH SUBSET (I.E. INDIVIDUAL
            # NIGHTS) ...
            for k, v in list(distinctmjds.items()):
                # GIVE ME THE MEAN MJD
                meanmjd = sum(v["mjds"]) / len(v["mjds"])
                summed_magnitudes[fil]["mjds"].append(meanmjd)
                # GIVE ME THE MEAN FLUX
                meanflux = sum(v["mags"]) / len(v["mags"])
                summed_magnitudes[fil]["mags"].append(meanflux)
                # GIVE ME THE COMBINED ERROR
                sum_of_squares = sum(x**2 for x in v["magErrs"])
                comberror = math.sqrt(sum_of_squares) / len(v["magErrs"])
                summed_magnitudes[fil]["magErrs"].append(comberror)
                # 5-sigma limits
                comb5siglimit = 23.9 - 2.5 * math.log10(5.0 * comberror)
                summed_magnitudes[fil]["lim5sig"].append(comb5siglimit)
                # GIVE ME NUMBER OF DATA POINTS COMBINED
                n = len(v["mjds"])
                summed_magnitudes[fil]["n"].append(n)
                alldata.append(
                    {
                        "mjd": meanmjd,
                        "uJy": meanflux,
                        "duJy": comberror,
                        "F": fil,
                        "n": n,
                        "mag5sig": comb5siglimit,
                    }
                )
        print("completed the ``stack_photometry`` method")

        return alldata

    """
    The following code was taken and modified for the purposes of this package from
    https://github.com/HC-Hwang/wise_light_curves/blob/master/wise_light_curves.py

    Original Authors:
    - Matthew Hill
    - Hsiang-Chih Hwang

    Update Author:
    - Noah Franz
    """

    @staticmethod
    def _get_by_position(ra, dec, radius=2.5):
        allwise_cat = "allwise_p3as_mep"
        neowise_cat = "neowiser_p1bs_psd"
        query_url = "http://irsa.ipac.caltech.edu/cgi-bin/Gator/nph-query"
        payload = {
            "catalog": allwise_cat,
            "spatial": "cone",
            "objstr": " ".join([str(ra), str(dec)]),
            "radius": str(radius),
            "radunits": "arcsec",
            "outfmt": "1",
        }
        r = requests.get(query_url, params=payload)
        allwise = ascii.read(r.text)
        payload = {
            "catalog": neowise_cat,
            "spatial": "cone",
            "objstr": " ".join([str(ra), str(dec)]),
            "radius": str(radius),
            "radunits": "arcsec",
            "outfmt": "1",
            "selcols": "ra,dec,sigra,sigdec,sigradec,glon,glat,elon,elat,w1mpro,w1sigmpro,w1snr,w1rchi2,w2mpro,w2sigmpro,w2snr,w2rchi2,rchi2,nb,na,w1sat,w2sat,satnum,cc_flags,det_bit,ph_qual,sso_flg,qual_frame,qi_fact,saa_sep,moon_masked,w1frtr,w2frtr,mjd,allwise_cntr,r_allwise,pa_allwise,n_allwise,w1mpro_allwise,w1sigmpro_allwise,w2mpro_allwise,w2sigmpro_allwise,w3mpro_allwise,w3sigmpro_allwise,w4mpro_allwise,w4sigmpro_allwise",  # noqa: E501
        }
        r = requests.get(query_url, params=payload)

        neowise = ascii.read(r.text, guess=False, format="ipac")

        return allwise, neowise

    @staticmethod
    def _download_single_data(
        name, ra, dec, root_path="ipac/", radius=2.5, overwrite=False
    ):
        # ra, dec: in degree
        # name, ra, dec = row['Name'], row['RAJ2000'], row['DEJ2000']
        # name = 'J' + ra + dec
        if root_path[-1] != "/":
            root_path += "/"
        if (
            not overwrite
            and os.path.isfile(root_path + name + "_allwise.ipac")
            and os.path.isfile(root_path + name + "_neowise.ipac")
        ):
            pass
        else:
            allwise, neowise = DataFinder._get_by_position(ra, dec, radius=radius)
            allwise.write(
                root_path + name + "_allwise.ipac", format="ascii.ipac", overwrite=True
            )
            neowise.write(
                root_path + name + "_neowise.ipac", format="ascii.ipac", overwrite=True
            )

    @staticmethod
    def _get_data_arrays(table, t, mag, magerr):
        """Get the time series from a potentially masked astropy table"""
        if table.masked:
            full_mask = table[t].mask | table[mag].mask | table[magerr].mask
            t = table[t].data
            mag = table[mag].data
            magerr = table[magerr].data

            t.mask = full_mask
            mag.mask = full_mask
            magerr.mask = full_mask

            return t.compressed(), mag.compressed(), magerr.compressed()

        else:
            return table[t].data, table[mag].data, table[magerr].data

    @staticmethod
    def _make_full_lightcurve(allwise, neowise, band):
        """band = 'w1', 'w2', 'w3', or 'w4'"""
        """Get a combined AllWISE and NEOWISE lightcurve from their Astropy tables"""

        if band not in ["w1", "w2", "w3", "w4"]:
            raise ValueError("band can only be w1, w2, w3, or w4")

        use_neowise = band in {"w1", "w2"}
        use_allwise = allwise is not None

        if use_neowise and use_allwise:
            t, m, e = DataFinder._get_data_arrays(
                allwise, "mjd", band + "mpro_ep", band + "sigmpro_ep"
            )
            t_n, m_n, e_n = DataFinder._get_data_arrays(
                neowise, "mjd", band + "mpro", band + "sigmpro"
            )
            t, m, e = (
                np.concatenate((t, t_n)),
                np.concatenate((m, m_n)),
                np.concatenate((e, e_n)),
            )

        elif use_neowise and not use_allwise:
            t, m, e = DataFinder._get_data_arrays(
                neowise, "mjd", band + "mpro", band + "sigmpro"
            )

        elif not use_neowise and use_allwise:
            t, m, e = DataFinder._get_data_arrays(
                allwise, "mjd", band + "mpro_ep", band + "sigmpro_ep"
            )

        else:
            raise Exception("No good allwise or neowise data!")

        t_index = t.argsort()
        t, m, e = map(lambda e: e[t_index], [t, m, e])

        return t, m, e

    @staticmethod
    def _make_full_lightcurve_multibands(allwise, neowise, bands=["w1", "w2"]):
        t, m, e = DataFinder._make_full_lightcurve(allwise, neowise, bands[0])
        filts = [bands[0] for i in range(len(t))]
        for band in bands[1:]:
            try:
                t_tmp, m_tmp, e_tmp = DataFinder._make_full_lightcurve(
                    allwise, neowise, band
                )
            except Exception:
                continue
            t = np.concatenate((t, t_tmp))
            m = np.concatenate((m, m_tmp))
            e = np.concatenate((e, e_tmp))
            filts += [band for i in range(len(t_tmp))]
        return t, m, e, np.array(filts)

    @staticmethod
    def _cntr_to_source_id(cntr):
        cntr = str(cntr)

        # fill leanding 0s
        if len(cntr) < 19:
            num_leading_zeros = 19 - len(cntr)
            cntr = "0" * num_leading_zeros + cntr

        pm = "p"
        if cntr[4] == "0":
            pm = "m"

        t = chr(96 + int(cntr[8:10]))

        return "%s%s%s_%cc%s-%s" % (
            cntr[0:4],
            pm,
            cntr[5:8],
            t,
            cntr[11:13],
            cntr[13:19],
        )

    @staticmethod
    def _only_good_data(allwise, neowise, verbose=False):
        """
        Select good-quality data. The criteria include:
        - matching the all-wise ID

        To be done:
        - deal with multiple cntr

        This filtering is described here:
        https://wise2.ipac.caltech.edu/docs/release/neowise/expsup/sec2_3.html
        """

        neowise_prefilter_n = len(neowise)
        neowise = neowise[
            (neowise["qual_frame"] > 0.0)
            * (neowise["qi_fact"] > 0.9)
            * (neowise["saa_sep"] > 0)
            * (neowise["moon_masked"] == "00")
        ]
        neowise_postfilter_n = len(neowise)
        if verbose:
            print(
                f"Filtered out {neowise_prefilter_n-neowise_postfilter_n} neowise \
                points, leaving {neowise_postfilter_n}"
            )

        cntr_list = []
        for data in neowise:
            if data["allwise_cntr"] not in cntr_list and data["allwise_cntr"] > 10.0:
                cntr_list.append(data["allwise_cntr"])

        if len(cntr_list) >= 2:
            print("multiple cntr:")
            print(cntr_list)
            return None, neowise

        if len(cntr_list) == 0:
            # import pdb; pdb.set_trace()
            # raise Exception('No center!')
            return None, neowise

        cntr = cntr_list[0]

        source_id = DataFinder._cntr_to_source_id(cntr)

        allwise_prefilter_n = len(allwise)
        allwise = allwise[
            (allwise["source_id_mf"] == source_id)
            * (allwise["saa_sep"] > 0.0)
            * (allwise["moon_masked"] == "0000")
            * (allwise["qi_fact"] > 0.9)
        ]
        allwise_postfilter_n = len(neowise)
        if verbose:
            print(
                f"Filtered out {allwise_prefilter_n-allwise_postfilter_n} allwise \
                points, leaving {allwise_postfilter_n}"
            )

        return allwise, neowise
